"""Provide utilities for helping color enhancement routines."""

from collections.abc import Sequence

from .shared import ColorSegment


def compress_color_data(
    color_data: list[ColorSegment],
) -> list[ColorSegment]:
    """
    Compress color data by removing duplicate entries.

    Parameters
    ----------
    color_data : list of tuple of float
        List of color segments.

    Returns
    -------
    list of tuple of float
        Compressed color table.
    """
    compressed: list[ColorSegment] = [color_data[0]]

    for x2, c2, d2 in color_data[1:]:
        x1, c1, _ = compressed[-1]
        if x1 == x2:
            compressed[-1] = (x1, c1, d2)
        else:
            compressed.append((x2, c2, d2))

    return compressed


def _interpv(x: float, x0: float, x1: float, y0: float, y1: float) -> float:
    slope = (y1 - y0) / (x1 - x0)
    return y0 + slope * (x - x0)


def _interp(x: float, xp: Sequence[float], yp: Sequence[float]) -> float:
    if x < xp[0] or x >= xp[-1]:
        raise ValueError("Value out of range")

    i = 0
    while i + 1 < len(xp) and x >= xp[i + 1]:
        i += 1

    x0, y0 = xp[i], yp[i]
    x1, y1 = xp[i + 1], yp[i + 1]

    return _interpv(x, x0, x1, y0, y1)


def interp(x: float, xp: Sequence[float], yp: Sequence[float]) -> float:
    """
    Interpolate x based on xp and yp using linear interpolation.

    Parameters
    ----------
    x : float
        The value to interpolate.
    xp : Sequence of float
        The x-coordinates of the data points.
    yp : Sequence of float
        The y-coordinates of the data points.

    Returns
    -------
    float
        The interpolated value.
    """
    return yp[-1] if x == xp[-1] else _interp(x, xp, yp)


def interpc(x: float, xp: Sequence[float], yp: Sequence[float]) -> float:
    """
    Interpolate x based on xp and yp using linear interpolation.

    Interpolate x based on xp and yp using linear interpolation with
    copying.

    Parameters
    ----------
    x : float
        The value to interpolate.
    xp : Sequence of float
        The x-coordinates of the data points.
    yp : Sequence of float
        The y-coordinates of the data points.

    Returns
    -------
    float
        The interpolated value.
    """
    if x <= xp[0]:
        return yp[0]

    return yp[-1] if x >= xp[-1] else _interp(x, xp, yp)


def interpx(x: float, xp: Sequence[float], yp: Sequence[float]) -> float:
    """
    Interpolate x based on xp and yp using linear interpolation.

    Interpolate x based on xp and yp using linear interpolation with
    extrapolation.

    Parameters
    ----------
    x : float
        The value to interpolate.
    xp : Sequence of float
        The x-coordinates of the data points.
    yp : Sequence of float
        The y-coordinates of the data points.

    Returns
    -------
    float
        The interpolated value.
    """
    if x < xp[0]:
        return _interpv(x, xp[0], xp[1], yp[0], yp[1])

    if x > xp[-1]:
        return _interpv(x, xp[-2], xp[-1], yp[-2], yp[-1])

    return interp(x, xp, yp)



from collections.abc import Sequence

from .shared import DomainData


class EnhancementColormap:

    cmap_names: Sequence[str]
    domain: DomainData
    extent: DomainData
    keypoints: Sequence[float]
    name: str

    def __init__(
        self,
        name: str,
        cmap_names: str | Sequence[str],
        keypoints: Sequence[float],
    ) -> None:
        if isinstance(cmap_names, str):
            cmap_names = [cmap_names]

        if len(keypoints) != len(cmap_names) + 1:
            raise ValueError(
                f"Expected {len(cmap_names) + 1} keypoints, got {len(keypoints)}"
            )

        for i in range(1, len(keypoints)):
            if keypoints[i] <= keypoints[i - 1]:
                raise ValueError("Keypoints must be monotonically increasing")

        self.cmap_names = cmap_names
        self.domain = keypoints[0], keypoints[-1]
        self.extent = self.domain
        self.keypoints = keypoints
        self.name = name



from datetime import datetime
from typing import Any, cast

from netCDF4 import Dataset  # pylint: disable=no-name-in-module
from numpy import datetime64, float64, nan, newaxis

from ..geodesy import RectangularRegion
from ..netcdf import DatasetView, HasStrHelp, scalar, variable
from ..utils.array import ArrayBool, ArrayFloat64, ArrayInt8, MaskedFloat32
from .metadata import DeltaTimeMetadata, TimeMetadata, VariableMetadata
from .netcdf_geodetic import GSLatLonGrid, LimitType

SECONDS_IN_DAY = 86400
SECONDS_IN_MINUTE = 60

NOT_A_DATETIME = cast(datetime, datetime64("NaT"))

MetadataType = TimeMetadata | DeltaTimeMetadata | VariableMetadata


class GSTimeData(HasStrHelp):

    # Delta-time data in minutes.
    delta_time: MaskedFloat32

    # Optimal-time (in days since 1970-01-01 UTC).
    optimal_time: float64

    # Optimal-time-bounds (in days since 1970-01-01 UTC).
    optimal_time_bounds: ArrayFloat64


class GSTimeGrid(GSTimeData):

    grid: GSLatLonGrid

    metadata: dict[str, MetadataType]

    def __init__(self, record: Dataset, grid: GSLatLonGrid) -> None:
        data = self._extract_timedata(record, grid.lon_limits, grid.lat_limits)

        self.grid = grid

        self.delta_time = data.delta_time
        self.optimal_time = data.optimal_time
        self.optimal_time_bounds = data.optimal_time_bounds

        self.metadata = self._get_metadata(record)

    @staticmethod
    def _extract_bounds_metadata(record: Dataset) -> VariableMetadata:
        coordinate = variable("time_bounds")

        class _TimeMetata(DatasetView):
            long_name: str = coordinate.attribute()
            comment: str = coordinate.attribute()
            units: str = coordinate.attribute()
            shape: tuple[int] = coordinate.attribute()

        metadata = _TimeMetata(record)

        return VariableMetadata(metadata)

    @staticmethod
    def _extract_delta_metadata(record: Dataset) -> DeltaTimeMetadata:
        measurement = variable("delta_time")

        class _ImageMetata(DatasetView):
            long_name: str = measurement.attribute()
            coordinates: str = measurement.attribute()
            units: str = measurement.attribute()
            range: ArrayInt8 = measurement.attribute("actual_range")
            content_type: str = measurement.attribute("coverage_content_type")
            shape: tuple[int] = measurement.attribute()

        metadata = _ImageMetata(record)

        return DeltaTimeMetadata(metadata)

    @staticmethod
    def _extract_time_metadata(record: Dataset) -> TimeMetadata:
        coordinate = variable("time")

        class _LatLonMetata(DatasetView):
            long_name: str = coordinate.attribute()
            comment: str = coordinate.attribute()
            standard_name: str = coordinate.attribute()
            units: str = coordinate.attribute()
            axis: str = coordinate.attribute()
            calendar: str = coordinate.attribute()
            content_type: str = coordinate.attribute("coverage_content_type")
            shape: tuple[int] = coordinate.attribute()

        metadata = _LatLonMetata(record)

        return TimeMetadata(metadata)

    @staticmethod
    def _extract_timedata(
        record: Dataset,
        lon_limits: LimitType,
        lat_limits: LimitType,
    ) -> "GSTimeData":
        def slice(x: Any) -> Any:
            min_lon, max_lon = lon_limits
            min_lat, max_lat = lat_limits
            return x[0, min_lat:max_lat, min_lon:max_lon]

        class _GSTimeData(DatasetView):
            delta_time: MaskedFloat32 = variable("delta_time").array(
                filter=slice
            )
            optimal_time: float64 = scalar("time")
            optimal_time_bounds: ArrayFloat64 = variable("time_bounds").data()

        data = _GSTimeData(record)

        data.delta_time.data[data.delta_time.mask] = nan

        return cast(GSTimeData, data)

    @classmethod
    def _get_metadata(cls, record: Dataset) -> dict[str, MetadataType]:
        return {
            "delta_time": cls._extract_delta_metadata(record),
            "time": cls._extract_time_metadata(record),
            "time_bounds": cls._extract_bounds_metadata(record),
        }

    @property
    def time(self) -> ArrayFloat64:
        # (in seconds since 1970-01-01 UTC)
        actual_time: ArrayFloat64 = (
            self.optimal_time * SECONDS_IN_DAY
            + self.delta_time.data * SECONDS_IN_MINUTE
        )
        return actual_time

    @property
    def time_bounds(self) -> ArrayFloat64:
        # (in seconds since 1970-01-01 UTC)
        actual_time_bounds: ArrayFloat64 = (
            self.optimal_time_bounds[:, newaxis].T * SECONDS_IN_DAY
            + self.delta_time.data * SECONDS_IN_MINUTE
        )
        return actual_time_bounds

    @property
    def mask(self) -> ArrayBool:
        return cast(ArrayBool, self.delta_time.mask)

    @property
    def region(self) -> RectangularRegion:
        return self.grid.region


class GSCoverageTime(HasStrHelp):

    datetime_start: datetime = NOT_A_DATETIME
    datetime_end: datetime = NOT_A_DATETIME

    def __init__(self, record: Dataset) -> None:
        datetime_start = getattr(record, "time_coverage_start", "")
        datetime_end = getattr(record, "time_coverage_end", "")

        if datetime_start and datetime_end:
            self.datetime_start = datetime.fromisoformat(datetime_start)
            self.datetime_end = datetime.fromisoformat(datetime_end)

    @property
    def timestamp_start(self) -> float:
        return self.datetime_start.timestamp()

    @property
    def timestamp_end(self) -> float:
        return self.datetime_end.timestamp()



"""
Provide locator for GOES-R Series imagery dataset's ABI products.

Classes:
    - GOESProductLocatorABI: All primary and derived Advanced Baseline
      Imager (ABI) products.
"""

from .locator import GOESProductLocator


class GOESProductLocatorABI(GOESProductLocator):
    """
    Product locator for GOES-R Series imagery dataset's ABI products.

    Instrument: Advanced Baseline Imager (ABI).
    Product: All primary and derived ABI products.
    """

    # Available scenes for the GOES-R Series imagery dataset products:
    #
    # NOTE: In its strictest sense, “Contiguous United States” refers
    # to the lower 48 states in North America (including the District
    # of Columbia), and “Continental United States” refers to 49 states
    # (including Alaska and the District of Columbia).
    AVAILABLE_SCENES: dict[str, str] = {
        GOESProductLocator.FULL_DISK: "Full Disk",
        GOESProductLocator.CONUS: "CONUS (Continental United States)",
        GOESProductLocator.MESO_1: "Mesoscale (Domain 1)",
        GOESProductLocator.MESO_2: "Mesoscale (Domain 2)",
    }

    # Instrument: Advanced Baseline Imager (ABI).
    INSTRUMENT_NAME: str = "ABI"

    # Available scan modes for the GOES-R Series imagery dataset ABI
    # products, regarding the requested scene for the product:
    # - Mode 3 (Previous Flex Mode)
    # - Mode 6 (Current Flex Mode)
    CM_MODES: list[str] = ["M3", "M6"]
    F_MODES: list[str] = ["M4"] + CM_MODES

    LEVEL_RAD: str = "L1b"
    LEVEL_NOT_RAD: str = "L2"
    DEFAULT_PRODUCT_LEVEL: str = LEVEL_NOT_RAD

    def __init__(
        self,
        name: str,
        level: str,
        scene: str,
        channels: list[str],
        origin: str,
    ) -> None:
        """
        Initialise a GOES-R Series imagery dataset ABI product locator.

        Constructs a GOES-R Series imagery dataset ABI product locator
        object.

        Parameters
        ----------
        name : str
            The name of the GOES-R Series imagery dataset ABI product.
            Due to how the dataset directories are organised, only a
            single product can be provided.
        level : str
            The level of the GOES-R Series imagery dataset product, e.g.
            "L1b" or "L2".
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        channels : list[str]
            The list of channels of the GOES-R Series imagery dataset
            ABI product, e.g. "C08" or "C13".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.

        Notes
        -----
        ValueError is raised if the provided origin, level or scene is
        invalid.
        """
        # TODO: Too many positional arguments. Solve it by using
        #       the Builder or Factory methods, or patterns like
        #       Essence or Fluent.
        self._validate_scene(scene, self.AVAILABLE_SCENES)

        scan_modes: list[str] = (
            self.F_MODES if scene == self.FULL_DISK else self.CM_MODES
        )

        super().__init__(
            name=name,
            level=level,
            scene=scene,
            instrument=self.INSTRUMENT_NAME,
            modes=scan_modes,
            channels=channels,
            origin=origin,
        )



import math

import cartopy.crs as ccrs
import numpy as np
from scipy.spatial import cKDTree


def calculate_image_size(
    extent_deg,
    resolution_m,
    target_crs,
    dpi,
):
    n_rows, n_cols, _, _ = calculate_matrix_size(
        extent_deg, resolution_m, target_crs
    )

    width_in, height_in = n_cols / dpi, n_rows / dpi

    return width_in, height_in


def calculate_matrix_size(
    extent_deg,
    resolution_m,
    target_crs,
):
    # Compute the extent size in metres
    width_m, height_m = get_extent_metre(extent_deg, target_crs)

    # Compute the grid size
    n_cols = 2 * (math.ceil(width_m / resolution_m) // 2)
    n_rows = 2 * (math.ceil(height_m / resolution_m) // 2)

    return n_rows, n_cols, height_m, width_m


def create_matrix(
    data,
    lat,
    lon,
    extent_deg,
    resolution_m,
    target_crs,
):
    n_rows, n_cols, height_m, width_m = calculate_matrix_size(
        extent_deg, resolution_m, target_crs
    )

    half_width_m = 0.5 * width_m
    half_height_m = 0.5 * height_m

    # Create the target grid in projected coordinates
    x_target = np.linspace(-half_width_m, half_width_m, n_cols)
    y_target = np.linspace(half_height_m, -half_height_m, n_rows)
    x_target_grid, y_target_grid = np.meshgrid(x_target, y_target)

    # Define the source projection
    source_crs = ccrs.PlateCarree(globe=target_crs.globe)

    # Transform the target grid coordinates to lat/lon
    points = source_crs.transform_points(
        target_crs, x_target_grid, y_target_grid
    )
    lon_grid, lat_grid = points[..., 0], points[..., 1]

    # Flatten the grids
    target_points = np.vstack((lon_grid.ravel(), lat_grid.ravel())).T
    source_points = np.vstack((lon.ravel(), lat.ravel())).T
    data_flat = data.ravel()

    # Build KDTree and interpolate
    tree = cKDTree(source_points)
    _, idx = tree.query(
        target_points, k=4
    )  # Get the 4 nearest neighbors for bilinear interpolation
    weights = 1 / np.maximum(
        tree.query(target_points, k=4)[0], 1e-12
    )  # Avoid division by zero
    weights /= np.sum(weights, axis=1, keepdims=True)  # Normalize weights

    # Compute interpolated values
    reprojected_data_flat = np.sum(data_flat[idx] * weights, axis=1)

    # Reshape to grid
    return reprojected_data_flat.reshape((n_rows, n_cols))



from .scale import EnhancementScale


def _ircolor() -> EnhancementScale:
    # Define the properties of a stock palette
    key_points = (200.0, 240.0, 320.0)  # (180, 240, 330)
    color_maps = ("jet_r", "gray_r")
    cmap_name = f"IRCOLOR {int(key_points[0])}-{int(key_points[1])}K"

    # Get a stock palette and configure the ticks for the colorbar, for this example
    return EnhancementScale.from_colormap(cmap_name, color_maps, key_points)


cmap: dict[str, EnhancementScale] = {"IRCOLOR": _ircolor()}



"""
Provide a class to manage HTTP request headers.

Defines the `RequestHeaders` class, which is used to manage HTTP request
headers for the GOES-DL project.

Classes:
    RequestHeaders: A class to represent HTTP headers for requests.
"""

import os
import platform

from ..info import __package_id__, __version__

IMAGE_JPEG: str = "image/jpeg"
IMAGE_PNG: str = "image/png"
TEXT_HTML: str = "text/html"
APPLICATION_JSON: str = "application/json"

# THREDDS defines two mime-types for NetCDF:
APPLICATION_NETCDF3: str = "application/x-netcdf"
APPLICATION_NETCDF4: str = "application/x-netcdf4"

ACCEPT_LANGUAGE: str = (
    "en-GB;q=0.9,en-US;q=0.8,en;q=0.7," + "es-ES;q=0.8,es-PY;q=0.7,es;q=0.6"
)


class RequestHeaders:
    """A class to represent HTTP headers for requests."""

    def __init__(self, *, accept: str | None = None) -> None:
        """
        Initialize the RequestHeaders object.

        Parameters
        ----------
        accept : str | None, optional
            The value of the "accept" header, by default TEXT_HTML
            ("text/html").
        """
        if accept is None:
            accept = TEXT_HTML
        self._headers: dict[str, str] = self._build_headers(accept)

    @staticmethod
    def get_user_agent() -> str:
        """
        Get the user agent string.

        Returns
        -------
        str
            The user agent string.
        """
        return (
            f"{__package_id__}/{__version__} "
            f"({platform.system()} {os.name.upper()} "
            f"{platform.release()}/{platform.version()})"
        )

    @staticmethod
    def _build_headers(accept: str) -> dict[str, str]:
        """
        Build the headers dictionary.

        Parameters
        ----------
        accept : str
            The value of the "accept" header.

        Returns
        -------
        dict[str, str]
            A dictionary containing the headers.
        """
        user_agent: str = RequestHeaders.get_user_agent()

        return {
            "accept": accept,
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": ACCEPT_LANGUAGE,
            "cache-control": "no-cache",
            "connection": "keep-alive",
            "pragma": "no-cache",
            "user-agent": user_agent,
        }

    @property
    def headers(self) -> dict[str, str]:
        """dict[str, str]: A dictionary containing the HTTP headers."""
        return self._headers



"""Provide constants for the enhancement utility module."""

BRG_MAX = 255.0
CLR_MAX = 255.0
CMYK_MAX = 100.0
HUE_MAX = 360.0
HSV_MAX = 100.0

CM_BGR = "BGR"
CM_RGB = "RGB"
CM_CMYK = "CMYK"
CM_GRAY = "GRAY"
CM_HSV = "HSV"

UNNAMED_TABLE = "UNNAMED"



"""
Provide the DatasourceAWS class for handling AWS-based data sources.

Classes:
    DatasourceAWS: Handle AWS-based data sources.
"""

from typing import TYPE_CHECKING, Literal
from urllib.parse import ParseResult

import boto3
from botocore import UNSIGNED
from botocore.client import ClientError, Config

from ..dataset import ProductLocator
from ..utils.url import URL
from .datasource_base import DatasourceBase
from .datasource_cache import DatasourceCache

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client

AWS_CLIENT: Literal["s3"] = "s3"


class DatasourceAWS(DatasourceBase):
    """
    Handle AWS-based data sources.

    The AWS S3 datasource is responsible for listing the contents of a
    directory in a remote location and for downloading files from that
    location. The base URL of the datasource is the URL of the AWS S3
    bucket.

    Methods
    -------
    download_file(file_path: str)
        Retrieve a file from the datasource and save it into the local
        repository.
    list_files(dir_path: str)
        List the contents of a remote directory.

    Attributes
    ----------
    bucket_name : str
        The name of the AWS S3 bucket.
    s3_client : S3Client
        The AWS S3 client.
    """

    bucket_name: str
    s3_client: "S3Client"

    def __init__(
        self,
        locator: ProductLocator | tuple[str, ...],
        cache: float | DatasourceCache | None = None,
    ) -> None:
        """
        Initialize the DatasourceAWS object.

        Parameters
        ----------
        locator : ProductLocator | tuple[str, ...]
            A `ProductLocator` object or a tuple of strings containing
            the base URL and an optional region where the S3 bucket is
            located. E.g. "us-west-1", "us-east-1", "eu-west-1", etc. If
            None, the default region is used.
        cache : float | DatasourceCache | None, optional
            The cache expiration time in seconds, by default None.

        Raises
        ------
        ValueError
            If the bucket does not exist or the user has no access.
        """
        base_url: str
        region: str | None
        if isinstance(locator, ProductLocator):
            (base_url, region) = locator.get_base_url("AWS")
        else:
            (base_url, region) = locator

        url_parts: ParseResult = URL.parse(base_url)

        bucket_name: str = url_parts.netloc

        self.s3_client: S3Client = self._get_client(region)

        if not self._bucket_exists(bucket_name):
            raise ValueError(
                f"Bucket '{bucket_name}' does not exist or you have no access."
            )

        super().__init__(base_url, cache)

        self.bucket_name: str = bucket_name

    def download_file(self, file_path: str) -> bytes:
        """
        Download a file from the datasource into the local repository.

        Get a file from a remote location or local repository. The path
        provided must be relative to the base URL and local repository
        root directory. The remote path is reconstructed in the local
        repository.

        Parameters
        ----------
        file_path : str
            The path to the remote file to be downloaded.

        Returns
        -------
        bytes
            The content of the file as a byte string.

        Raises
        ------
        RuntimeError
            If the file cannot be retrieved.
        """
        try:
            return self._retrieve_file(file_path)

        except ClientError as exc:
            raise RuntimeError(
                f"Unable to retrieve the file '{file_path}': {exc}"
            ) from exc

    def list_files(self, dir_path: str) -> list[str]:
        """
        List the contents of a directory.

        List the contents of a directory in a remote location. The path
        is relative to the base URL.

        Parameters
        ----------
        dir_path : str
            The path to the directory. The path is relative to the base
            URL.

        Returns
        -------
        list[str]
            A list of file names in the directory.
        """
        cached_list = self.cache.get_item(dir_path)

        if cached_list is not None:
            return cached_list

        folder_path = self._get_item_path(dir_path)

        paginator = self.s3_client.get_paginator("list_objects_v2")
        pages = paginator.paginate(Bucket=self.bucket_name, Prefix=folder_path)

        # Workaround for non-existing folders.
        for page in pages:
            if page["KeyCount"] == 0:
                return []

            break

        ss = len(folder_path)

        file_list: list[str] = []

        file_list.extend(
            f"{dir_path}{obj['Key'][ss:]}"
            for page in pages
            for obj in page["Contents"]
            if obj["Size"] > 0
        )

        self.cache.add_item(dir_path, file_list)

        return file_list

    def _bucket_exists(self, bucket_name: str) -> bool:
        try:
            self.s3_client.head_bucket(Bucket=bucket_name)

        except ClientError:  # as exc
            # Bucket not found error or host is out of service (log this!)
            return False

        return True

    @staticmethod
    def _get_client(region: str | None) -> "S3Client":
        if region:
            return boto3.client(
                AWS_CLIENT,
                region_name=region,
                config=Config(signature_version=UNSIGNED),
            )

        return boto3.client(
            AWS_CLIENT,
            config=Config(signature_version=UNSIGNED),
        )

    def _get_item_path(self, dir_path: str) -> str:

        folder_url = self._url_join(self.base_url, dir_path)

        url_parts = URL.parse(folder_url)

        return url_parts.path[1:]

    def _object_exists(self, bucket_name: str, object_path: str) -> bool:
        try:
            self.s3_client.head_object(Bucket=bucket_name, Key=object_path)

        except ClientError:  # as exc
            # Object not found error or host is out of service (log this!)
            return False

        return True

    def _retrieve_file(self, file_path: str) -> bytes:
        folder_path = self._get_item_path(file_path)

        response = self.s3_client.get_object(
            Bucket=self.bucket_name, Key=folder_path
        )

        return response["Body"].read()

    @staticmethod
    def _url_join(head: str, tail: str) -> str:
        # Note: url.join() fails with "s3://" URLs.
        # > folder_url: str = url.join(self.base_url, dir_path)

        if head.endswith("/") and tail.startswith("/"):
            head = head[:-1]

        if not head.endswith("/") and not tail.startswith("/"):
            return f"{head}/{tail}"

        return head + tail



"""
Provide a caching mechanism for directory file lists.

Classes:
    DatasourceCacheItem: A data class representing a cached item.
    DatasourceCache: A class for caching the file list of a directory.
"""

import time
from dataclasses import dataclass


@dataclass(frozen=True)
class DatasourceCacheItem:
    """
    Data class for a cached datasource item.

    Represent an item in the cache. The item contains the list of files
    in a directory and the time when the cache was created.

    Attributes
    ----------
    files : list[str]
        The list of files in the directory.
    created_at : float
        The time when the cache was created.
    """

    files: list[str]
    created_at: float


class DatasourceCache:
    """
    Cache for the file list of a datasource directory/bucket.

    Implement a caching mechanism for the file list of a directory.  The
    cache is used to store the list of files in a directory for a
    certain amount of time. The cache is cleared every time the time
    life is reached.

    Methods
    -------
    add_item(dir_path: str, files: list[str]) -> None
        Add a list of files to the cache.
    clean_cache(all_items: bool = False) -> None
        Clean the cache.
    get_item(dir_path: str) -> list[str] | None
        Get the list of files in a directory.
    has_item(dir_path: str) -> bool
        Check if a directory is in the cache.
    remove_item(dir_path: str = "") -> None
        Clear the cache.

    Notes
    -----
    Set `life_time` to a positive number of second to enable the cache.
    The cache will be cleared every time the `life_time` is reached.
    Setting `life_time` to a non-positive number will disable the cache.
    If `life_time` is set to "nan", the behavior is undefined. The
    default value is +inf.
    """

    def __init__(self, life_time: float | None = None) -> None:
        """
        Initialize the cache.

        Initialize the cache with an optional life time.

        Parameters
        ----------
        life_time : float | None, optional
            The time in seconds that an item will be kept in the cache.
            (default: +inf)
        """
        self.life_time: float = (
            float("+inf") if life_time is None else life_time
        )
        self.cache: dict[str, DatasourceCacheItem] = {}

    def add_item(self, dir_path: str, files: list[str]) -> None:
        """
        Add a list of files to the cache.

        Add a list of files to the cache. The list of files is
        associated with the directory path.

        Parameters
        ----------
        dir_path : str
            The path to the directory. The path is relative to the base
            URL.
        files : list[str]
            The list of files in the directory.

        Raises
        ------
        ValueError
            If the folder is already in the cache.
        """
        if dir_path in self.cache:
            raise ValueError(f"Folder '{dir_path}' already in cache.")

        current_time: float = time.time()

        cache_item: DatasourceCacheItem = DatasourceCacheItem(
            files, current_time
        )
        self.cache[dir_path] = cache_item

    def clean_cache(self, all_items: bool = False) -> None:
        """
        Clean the cache.

        Clean the cache by removing items that have expired. If
        `all_items` is True, all items in the cache will be removed.

        Parameters
        ----------
        all_items : bool, optional
            If True, all items in the cache will be removed. If False,
            only the items that have expired will be removed. (default:
            False)
        """
        if all_items:
            self.cache.clear()
            return

        current_time: float = time.time()

        for dir_path, cache_item in list(self.cache.items()):
            expire_time: float = cache_item.created_at + self.life_time

            if expire_time < current_time:
                self.cache.pop(dir_path)

    def get_item(self, dir_path: str) -> list[str] | None:
        """
        Get the list of files in a directory.

        Get the list of files in a directory from the cache.

        Parameters
        ----------
        dir_path : str
            The path to the directory. The path is relative to
            the base URL.

        Returns
        -------
        list[str] | None
            The list of files in the directory or None if the
            directory is not found in the cache.
        """
        if dir_path in self.cache:
            cache_item: DatasourceCacheItem = self.cache[dir_path]

            expire_time: float = cache_item.created_at + self.life_time
            current_time: float = time.time()

            if expire_time > current_time:
                return cache_item.files

            self.remove_item(dir_path)

        return None

    def has_item(self, dir_path: str) -> bool:
        """
        Check if a directory is in the cache.

        Check if a directory is in the cache.

        Parameters
        ----------
        dir_path : str
            The path to the directory. The path is relative to
            the base URL.

        Returns
        -------
        bool
            True if the directory is in the cache, False otherwise.
        """
        return dir_path in self.cache

    def remove_item(self, dir_path: str = "") -> None:
        """
        Clear the cache.

        Clear the file list cache of the datasource object.

        Parameters
        ----------
        dir_path : str, optional
            The path to the directory. The path is relative to
            the base URL. If no path is provided, the entire
            cache should be cleared. (default: "")

        Raises
        ------
        ValueError
            If the folder is not found in the cache.
        """
        if not dir_path:
            raise ValueError(f"Folder '{dir_path}' not found in cache.")

        if dir_path in self.cache:
            self.cache.pop(dir_path, None)
            return

        self.cache.clear()



"""
Provide the DatasourceLocal class for handling local-based data sources.

Classes:
    DatasourceLocal: Handle local-based data sources.
"""

import contextlib
import os
from pathlib import Path

from ..utils import FileRepository
from .datasource_base import DatasourceBase
from .datasource_cache import DatasourceCache


class DatasourceLocal(DatasourceBase):
    """
    Handle local-based data sources.

    Provide methods to interact with local folders and files.

    Methods
    -------
    download_file(file_path: str)
        Retrieve a file from the datasource.
    list_files(dir_path: str)
        List the contents of a remote directory.

    Attributes
    ----------
    source : FileRepository
        A repository to manage files in a local drive
    """

    source: FileRepository

    def __init__(
        self,
        root_path: str | Path,
        cache: float | DatasourceCache | None = None,
    ) -> None:
        """
        Initialize the DatasourceLocal object.

        Parameters
        ----------
        root_path : str | Path
            The root path of a local-based data source.
        cache : float | DatasourceCache | None, optional
            The cache expiration time in seconds, by default None.

        Raises
        ------
        ValueError
            If the resource does not exist or the user has no access.
        """
        if not os.path.isdir(root_path):
            raise ValueError(
                f"Path '{root_path}' does not exist or you have no access."
            )

        self.source = FileRepository(root_path)

        super().__init__(str(root_path), cache)

    def download_file(self, file_path: str) -> bytes:
        """
        Download a file from the datasource into the local repository.

        Get a file from a remote location or local repository. The path
        provided must be relative to the base URL and local repository
        root directory. The remote path is reconstructed in the local
        repository.

        Parameters
        ----------
        file_path : str
            The path to the remote file to be downloaded.

        Returns
        -------
        bytes
            The content of the file as a byte string.

        Raises
        ------
        RuntimeError
            If the file cannot be retrieved or does not exist.
        """
        try:
            return self.source.read_file(file_path)

        except FileNotFoundError as exc:
            raise RuntimeError(
                f"The file '{file_path}' does not exist: {exc}"
            ) from exc
        except IOError as exc:
            raise RuntimeError(
                f"Unable to retrieve the file '{file_path}': {exc}"
            ) from exc

    def list_files(self, dir_path: str) -> list[str]:
        """
        List the contents of a directory.

        Lists files within a directory in a local drive and its
        subdirectories. The path is relative to the root path.

        Parameters
        ----------
        dir_path : str
            The path to the directory. The path is relative to the root
            path.

        Returns
        -------
        list[str]
            A list of file names in the directory.
        """
        cached_paths = self.cache.get_item(dir_path)

        if cached_paths is not None:
            return cached_paths

        folder_content: list[str] = []

        with contextlib.suppress(FileNotFoundError):
            folder_content = self.source.list_files(dir_path)

        folder_content = [os.path.join(dir_path, f) for f in folder_content]

        self.cache.add_item(dir_path, folder_content)

        return folder_content



from typing import cast

from cartopy import crs as ccrs
from cartopy.crs import Globe, PlateCarree, Projection
from cartopy.feature import NaturalEarthFeature
from cartopy.mpl.ticker import LatitudeFormatter, LongitudeFormatter
from matplotlib import pyplot as plt
from matplotlib.collections import QuadMesh
from matplotlib.ticker import MultipleLocator
from netCDF4 import Dataset

from ..enhancement import EnhancementScale, cmap
from ..geodesy import RectangularRegion
from .netcdf_geodetic import GSLatLonGrid
from .netcdf_image import GSImage
from .netcdf_metadata import GSDatasetMetadata
from .netcdf_time import GSCoverageTime

# Create the Natural Earth projection (Platé-Carrée projection on WGS84 ellipsoid)


def read_gridsat_dataset(
    dataframe: Dataset, channel: str, region: RectangularRegion
) -> tuple[GSImage, GSCoverageTime, GSDatasetMetadata]:
    grid = GSLatLonGrid(dataframe, region)  # , corners=True

    data = GSImage(dataframe, channel, grid)

    coverage = GSCoverageTime(dataframe)

    metadata = GSDatasetMetadata(dataframe, channel)

    return data, coverage, metadata


class GSPlotParameter:

    title: str | tuple[str, str] = ""
    axis_label: tuple[str, str] = "", ""
    cbar_label: str = ""

    def __init__(
        self,
        title: str | tuple[str, str] = "",
        axis_label: tuple[str, str] = ("", ""),
        cbar_label: str = "",
    ) -> None:
        self.title = title
        self.axis_label = axis_label
        self.cbar_label = cbar_label


class GSPlot:

    # The figure dimensions (in inches)
    fig_size: tuple[float, float] = 4.0, 4.0

    # The figure resolution (in number of dots per inch)
    fig_dpi: int = 200

    # The image output resolution (in number of dots per inch)
    img_dpi: int = 200

    near_earth_scale: str = "10m"

    enhancement: EnhancementScale

    crs: Projection

    def __init__(self) -> None:
        self.enhancement = cmap["IRCOLOR"]

        # Create the target plot projection (same as above)

        target_globe = Globe(ellipse="WGS84")

        self.crs = PlateCarree(central_longitude=0.0, globe=target_globe)

    def plot(
        self,
        image: GSImage,
        param: GSPlotParameter,
        save_path: str = "",
        show: bool = True,
    ) -> None:
        # Create the figure and setup the axes
        fig = plt.figure("map", figsize=self.fig_size, dpi=self.fig_dpi)

        # Create the axes ith the required projection, i.e., `target_crs` (see definition above)
        ax = fig.add_axes((0.1, 0.16, 0.80, 0.75), projection=self.crs)

        self._add_grid(ax, image.region, param.axis_label)

        self._add_admin_info(ax)

        self._add_crosshair(ax)

        # Plot the data (in `gcrs`, see definition above) with a color map from the stock
        # shading;
        #   - center:  lon.shape == image.shape => "nearest" == "auto" == None | "gouraud"
        #   - corners: lon.shape != image.shape => "flat" == "auto" == None
        mesh = self._plot_data(ax, image)

        self._add_colorbar(mesh, fig, param.cbar_label)

        self._add_title(ax, param.title)

        # Save the media file (ensure the destination path does exist)
        if save_path:
            plt.savefig(save_path, dpi=self.img_dpi, bbox_inches="tight")

        # Show the plot
        if show:
            plt.show()
        else:
            plt.close()

    def save(
        self, save_path: str, image: GSImage, param: GSPlotParameter
    ) -> None:
        self.plot(image, param, save_path, False)

    def show(self, image: GSImage, param: GSPlotParameter) -> None:
        self.plot(image, param)

    def _add_admin_info(self, ax: plt.Axes) -> None:  # type: ignore
        # Create the Natural Earth projection (Platé-Carrée projection
        # on WGS84 ellipsoid)

        natearth_globe = ccrs.Globe(ellipse="WGS84")

        natearth_crs = ccrs.PlateCarree(
            central_longitude=0.0,
            globe=natearth_globe,
        )

        # Create political boundaries, in `natearth_crs` (see definition above)
        provinces = NaturalEarthFeature(
            category="cultural",
            name="admin_1_states_provinces",
            scale=self.near_earth_scale,
            facecolor="none",
            transform=natearth_crs,
        )

        countries = NaturalEarthFeature(
            category="cultural",
            name="admin_0_countries",
            scale=self.near_earth_scale,
            facecolor="none",
            transform=natearth_crs,
        )

        # Add political boundaries to the plot using our plot projection
        ax.add_feature(
            provinces, edgecolor="gray", linewidth=0.25, transform=self.crs
        )
        ax.add_feature(
            countries, edgecolor="black", linewidth=0.25, transform=self.crs
        )

    def _add_colorbar(
        self, mesh: QuadMesh, fig: plt.Figure, label: str  # type: ignore
    ) -> None:

        # Add the colorbar
        caxes = fig.add_axes([0.12, 0.05, 0.76, 0.02])
        cb = plt.colorbar(
            mesh,
            ticks=self.enhancement.cticks,
            orientation="horizontal",
            extend="both",
            cax=caxes,
        )

        # Create a minor tick locator for the colorbar
        minor_locator = MultipleLocator(5)

        # Set the colorbar tick characteristics
        cb.ax.xaxis.set_minor_locator(minor_locator)
        cb.ax.tick_params(
            labelsize=4,
            labelcolor="black",
            width=0.5,
            length=2.0,
            direction="out",
            pad=1.0,
        )
        cb.ax.tick_params(axis="both", which="minor", length=1.5, width=0.4)

        # Set the colorbar caption
        # f"{cmi_entity} {cmi_measure} [{cmi_units}]" -> TOA Brightness Temperature
        # cb_label = f"{image.metadata.long_name} @ {metadata.wavelength:.1f} μm [{image.metadata.units}]"
        cb.set_label(label=label, size=5.0, color="black", weight="normal")

        # Set the colorbar characteristics
        cb.outline.set_linewidth(0.4)  # type: ignore[operator]

    def _add_crosshair(self, ax: plt.Axes) -> None:  # type: ignore
        # Add a crosshair to the plot
        xlims = ax.get_xlim()
        ylims = ax.get_ylim()
        center_x = (xlims[1] + xlims[0]) / 2
        center_y = (ylims[1] + ylims[0]) / 2
        ax.axvline(x=center_x, color="red", linewidth=0.4)
        ax.axhline(y=center_y, color="red", linewidth=0.4)

    def _add_grid(
        self,
        ax: plt.Axes,  # type: ignore
        region: RectangularRegion,
        labels: tuple[str, str],
    ) -> None:

        # Set gridline ticks characteristics
        ax.tick_params(
            left=True,
            right=True,
            bottom=True,
            top=True,
            labelleft=True,
            labelright=False,
            labelbottom=True,
            labeltop=False,
            length=0.0,
            width=0.05,
            labelsize=4.5,
            labelcolor="black",
        )

        # Add gridlines, set the longitude and latitude grid major tick locations,
        # set gridline characteristics, and configure gridline labels
        ax.gridlines(
            xlocs=region.xticks,
            ylocs=region.yticks,
            linewidth=0.25,
            linestyle="--",
            color="red",
            alpha=0.6,
            draw_labels=False,
        )

        # Set X-axis label characteristics
        ax.set_xticks(region.xticks, crs=self.crs)
        ax.xaxis.set_major_formatter(
            LongitudeFormatter(dateline_direction_label=True)
        )
        ax.set_xlabel(
            labels[0],
            color="black",
            fontsize=6,
            labelpad=3.0,
        )

        # Set Y-axis label characteristics
        ax.set_yticks(region.yticks, crs=self.crs)
        ax.yaxis.set_major_formatter(LatitudeFormatter())
        ax.set_ylabel(
            labels[1],
            color="black",
            fontsize=6,
            labelpad=3.0,
        )

        # Set the map limits, in `target_crs` (see definition above)
        ax.set_extent(region.extent, crs=self.crs)

    def _add_title(
        self,
        ax: plt.Axes,  # type: ignore
        title: str | tuple[str, str],
    ) -> None:
        # Set the title
        if isinstance(title, str):
            plt.title(title)
            return

        ax.set_title(title[0], fontsize=6.5, loc="left")
        ax.set_title(title[1], fontsize=6.5, loc="right")

    def _plot_data(
        self,
        ax: plt.Axes,  # type: ignore
        data: GSImage,
    ) -> QuadMesh:
        # Plot the data (in `gcrs`, see definition above) with a color map from the stock
        # shading;
        #   - center:  lon.shape == image.shape => "nearest" == "auto" == None | "gouraud"
        #   - corners: lon.shape != image.shape => "flat" == "auto" == None

        mesh = ax.pcolormesh(
            data.grid.lon,
            data.grid.lat,
            data.image,
            shading="gouraud",  #
            cmap=self.enhancement.cmap,
            norm=self.enhancement.cnorm,
            transform=data.grid.crs,
        )

        return cast(QuadMesh, mesh)



import math

from pyproj import CRS, Transformer


def get_semimajor_axis(projection):
    # Get the ellipsoid semimajor axis in metres
    if projection.globe is None:
        ellipsoid = CRS("WGS84").ellipsoid
        return ellipsoid.semi_major_metre

    if projection.globe.semimajor_axis is None:
        globe = projection.globe
        ellipse_name = globe.ellipse or globe.datum
        ellipsoid = CRS(ellipse_name).ellipsoid
        return ellipsoid.semi_major_metre

    return projection.globe.semimajor_axis


def get_extent_metre(extent_deg, projection):
    # Get the ellipsoid semimajor axis in metres
    semimajor_axis = get_semimajor_axis(projection)

    # Compute the extent size in metres
    deg_to_m = math.pi * semimajor_axis / 180.0
    width_deg, height_deg = extent_deg
    width_m = width_deg * deg_to_m
    height_m = height_deg * deg_to_m

    return width_m, height_m


def get_centered_domain(extent_deg, aeqd, pc):
    width_m, height_m = get_extent_metre(extent_deg, aeqd)

    half_width_m = 0.5 * width_m
    half_height_m = 0.5 * height_m

    limits = [-half_width_m, half_width_m, -half_height_m, half_height_m]

    x_coords = [
        -half_width_m,
        -half_width_m,
        -half_width_m,
        0,
        0,
        half_width_m,
        half_width_m,
        half_width_m,
    ]
    y_coords = [
        -half_height_m,
        0,
        half_height_m,
        -half_height_m,
        half_height_m,
        -half_height_m,
        0,
        half_height_m,
    ]

    # lons_deg, lats_deg = transform(aeqd, pc, x_coords, y_coords)
    transformer = Transformer.from_crs(aeqd, pc, always_xy=True)
    lons_deg, lats_deg = transformer.transform(x_coords, y_coords)

    domain = [min(lons_deg), max(lons_deg), min(lats_deg), max(lats_deg)]

    return domain, limits



"""
Constants for the GridSat dataset file handling.

This module contains constants used for processing GridSat dataset
files, including file suffixes, date formats, and product versions.
"""

GRIDSAT_FILE_SUFFIX: str = ".nc"

B1_FILE_DATE_FORMAT: str = "%Y.%m.%d.%H"
B1_FILE_DATE_PATTERN: str = r"\d{4}\.\d{2}\.\d{2}\.\d{2}"
B1_FILE_PREFIX: str = "GRIDSAT"
B1_PATH_DATE_FORMAT: str = "%Y"
B1_PATH_PREFIX: str = ""
B1_PRODUCT_LATEST_VERSION: str = "v02r01"
B1_PRODUCT_NAME: str = "B1"

GOES_FILE_DATE_FORMAT: str = "%Y.%m.%d.%H%M"
GOES_FILE_DATE_PATTERN: str = r"\d{4}\.\d{2}\.\d{2}\.\d{4}"
GOES_FILE_PREFIX: str = "GridSat"
GOES_PATH_DATE_FORMAT: str = "%Y/%m"
GOES_PRODUCT_LATEST_VERSION: str = "v01"

NA = "not available"



from datetime import datetime


def iso_to_timestamp(iso_strings):
    """Convierte una lista de cadenas ISO a una lista de timestamps en segundos desde la época."""
    # Convertir fechas ISO a timestamps en segundos desde la época
    return [
        datetime.fromisoformat(iso_string).timestamp()
        for iso_string in iso_strings
    ]



"""
Provide locator for GOES-R Series imagery dataset's ABI products.

Classes:
    - GOESProductLocatorABINSO: All derived ABI products with name,
      scene, and origin.
"""

from .locator_abi import GOESProductLocatorABI


class GOESProductLocatorABINSO(GOESProductLocatorABI):
    """
    Product locator for GOES-R Series imagery dataset's ABI products.

    Instrument: Advanced Baseline Imager (ABI)
    Product: Multi-band primary and derived ABI products.
    """

    def __init__(self, name: str, scene: str, origin: str) -> None:
        """
        Initialise a GOES-R Series imagery dataset ABI product locator.

        Constructs a GOES-R Series imagery dataset ABI derived product
        locator object.

        Parameters
        ----------
        name : str
            The name of the GOES-R Series imagery dataset ABI product.
            Due to how the dataset directories are organised, only a
            single product can be provided.
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.
        """
        super().__init__(
            name=name,
            level=self.DEFAULT_PRODUCT_LEVEL,
            scene=scene,
            channels=[],
            origin=origin,
        )



import glob
import os
import pickle
from pathlib import Path


def load_dataset_list(
    root_path: Path | str, pattern: str = "*.nc"
) -> list[str]:
    """
    Lists files within a directory and its subdirectories.

    Parameters
    ----------
    root_path : Path | str
        The root path of the directory to explore.

    Returns
    -------
    list[str]
        A list with the relative path of all files within the directory.
    """
    # Recursivele retrieve the content of the root path
    content = glob.glob(os.path.join(root_path, "**", pattern), recursive=True)

    # Filtering the directory content to extract only the file elements
    files = filter(os.path.isfile, content)

    # Get the relative path
    return [os.path.relpath(file, root_path) for file in files]


def save_metadata(path, metadata):
    with open(path, "wb") as file:
        pickle.dump(metadata, file)


def load_metadata(path):
    with open(path, "rb") as file:
        return pickle.load(file)



from numpy import float64


class GeosProjParameters:
    """
    A class to store the geostationary projection parameters.

    Attributes
    ----------
    longitude_of_projection_origin : float64
        The longitude of the projection origin in East degrees.
    perspective_point_height : float64
        The height of the perspective point in meters.
    sweep_angle_axis : str
        The axis of the sweep angle.
    """

    # Information about the projection
    longitude_of_projection_origin: float64
    perspective_point_height: float64
    sweep_angle_axis: str

    def __init__(
        self,
        longitude_of_projection_origin: float64,
        perspective_point_height: float64,
        sweep_angle_axis: str,
    ) -> None:
        """
        Initialize the GeostationaryParameters object.

        Parameters
        ----------
        longitude_of_projection_origin : float64
            The longitude of the projection origin in East degrees.
        perspective_point_height : float64
            The height of the perspective point in meters.
        sweep_angle_axis : str
            The axis of the sweep angle.
        """
        self.longitude_of_projection_origin = longitude_of_projection_origin
        self.perspective_point_height = perspective_point_height
        self.sweep_angle_axis = sweep_angle_axis



import math

# Dataset name

dataset_name_gridsat_gc: str = "GridSat-GOES"

# Spatial resolution (GRS80 ellipsoid)

GRS80_EQUATORIAL_RADIUS_M = 6378137.0
GRS80_INVERSE_FLATTENING = 298.257223563
GRS80_EQUATORIAL_PERIMETER_M = 2.0 * math.pi * GRS80_EQUATORIAL_RADIUS_M
DEG_TO_KM = GRS80_EQUATORIAL_PERIMETER_M / 360000.0
PIXELS_PER_DEGREE = 25
SPATIAL_RESOLUTION_DEG = 1.0 / PIXELS_PER_DEGREE
SPATIAL_RESOLUTION_KM = SPATIAL_RESOLUTION_DEG * DEG_TO_KM

geospatial_resolution_deg = SPATIAL_RESOLUTION_DEG, SPATIAL_RESOLUTION_DEG
geospatial_resolution_km = SPATIAL_RESOLUTION_KM, SPATIAL_RESOLUTION_KM

# Dataset abstract

ppd = PIXELS_PER_DEGREE
dpc = SPATIAL_RESOLUTION_DEG
kpc = SPATIAL_RESOLUTION_KM

abstract_gridsat_gc = (
    f"This product is referred to as {dataset_name_gridsat_gc}. The "
    f"resolution of the grid is 1/{ppd}th of a degree, or {dpc:.2f} "
    f"degrees, equivalent to {kpc:.2f} km at the Equator, latitude "
    f"and longitude, yielding {ppd} pixels per degree."
)

# Long platform name

platform_origin_gridsat_gc = {
    f"GOES-{id}": f"G{id:0>2}" for id in range(8, 16)
}

origin_platform_gridsat_gc = {
    value: key for key, value in platform_origin_gridsat_gc.items()
}

# Channel description

channel_description_gc = {
    "ch1": "Reflectance (or Scaled Radiance) for the 0.6 µm band",
    "ch2": "Brightness Temperature of the 3.9 µm channel",
    "ch3": "Brightness Temperature of the 6.7 µm channel",
    "ch4": "Brightness Temperature of the 10.7 µm channel",
    "ch5": "Brightness Temperature of the 12.0 µm channel",
    "ch6": "Brightness Temperature of the 13.4 µm channel",
}

# GridSat-GOES channel to actual channel correspondence

channel_correspondence_il = {
    "ch1": 1,
    "ch2": 2,
    "ch3": 3,
    "ch4": 4,
    "ch5": 5,
    "ch6": 0,
}

channel_correspondence_mp = {
    "ch1": 1,
    "ch2": 2,
    "ch3": 5,
    "ch4": 4,
    "ch5": 0,
    "ch6": 3,
}

channel_correspondence = {
    "G08": channel_correspondence_il,
    "G09": channel_correspondence_il,
    "G10": channel_correspondence_il,
    "G11": channel_correspondence_il,
    "G12": channel_correspondence_mp,
    "G13": channel_correspondence_mp,
    "G14": channel_correspondence_mp,
    "G15": channel_correspondence_mp,
}

# Wavelength range lower bound in µm

wavelength_range_lower_bound_il = {
    1: 0.55,
    2: 3.80,
    3: 6.50,
    4: 10.20,
    5: 11.50,
}

wavelength_range_lower_bound_m = {
    1: 0.55,
    2: 3.80,
    3: 13.00,
    4: 10.20,
    5: 5.80,
}

wavelength_range_lower_bound_np = {
    1: 0.52,
    2: 3.73,
    3: 13.00,
    4: 10.20,
    5: 5.80,
}

wavelength_range_lower_bound = {
    "G08": wavelength_range_lower_bound_il,
    "G09": wavelength_range_lower_bound_il,
    "G10": wavelength_range_lower_bound_il,
    "G11": wavelength_range_lower_bound_il,
    "G12": wavelength_range_lower_bound_m,
    "G13": wavelength_range_lower_bound_np,
    "G14": wavelength_range_lower_bound_np,
    "G15": wavelength_range_lower_bound_np,
}

# Wavelength upper bound in µm

wavelength_range_upper_bound_il = {
    1: 0.75,
    2: 4.00,
    3: 7.00,
    4: 11.20,
    5: 12.50,
}

wavelength_range_upper_bound_m = {
    1: 0.75,
    2: 4.00,
    3: 13.70,
    4: 11.20,
    5: 7.30,
}

wavelength_range_upper_bound_np = {
    1: 0.71,
    2: 4.07,
    3: 13.70,
    4: 11.20,
    5: 7.30,
}

wavelength_range_upper_bound = {
    "G08": wavelength_range_upper_bound_il,
    "G09": wavelength_range_upper_bound_il,
    "G10": wavelength_range_upper_bound_il,
    "G11": wavelength_range_upper_bound_il,
    "G12": wavelength_range_upper_bound_m,
    "G13": wavelength_range_upper_bound_np,
    "G14": wavelength_range_upper_bound_np,
    "G15": wavelength_range_upper_bound_np,
}

# Measurement range lower bound in % albedo (1) / K (2-5)

measurement_range_lower_bound_im = {
    1: 1.6,
    2: 4.0,
    3: 4.0,
    4: 4.0,
    5: 4.0,
}

measurement_range_lower_bound_np = {
    1: 0.0,
    2: 4.0,
    3: 4.0,
    4: 4.0,
    5: 4.0,
}

measurement_range_lower_bound = {
    "G08": measurement_range_lower_bound_im,
    "G09": measurement_range_lower_bound_im,
    "G10": measurement_range_lower_bound_im,
    "G11": measurement_range_lower_bound_im,
    "G12": measurement_range_lower_bound_im,
    "G13": measurement_range_lower_bound_np,
    "G14": measurement_range_lower_bound_np,
    "G15": measurement_range_lower_bound_np,
}

# Measurement range upper bound in % albedo (1) / K (2-5)

measurement_range_upper_bound_ik = {
    1: 100.0,
    2: 320.0,
    3: 320.0,
    4: 320.0,
    5: 320.0,
}

measurement_range_upper_bound_lp = {
    1: 100.0,
    2: 335.0,
    3: 320.0,
    4: 320.0,
    5: 320.0,
}

measurement_range_upper_bound = {
    "G08": measurement_range_upper_bound_ik,
    "G09": measurement_range_upper_bound_ik,
    "G10": measurement_range_upper_bound_ik,
    "G11": measurement_range_upper_bound_lp,
    "G12": measurement_range_upper_bound_lp,
    "G13": measurement_range_upper_bound_lp,
    "G14": measurement_range_upper_bound_lp,
    "G15": measurement_range_upper_bound_lp,
}

# Measurement units

ALBEDO = "% albedo"
KELVIN = "Kelvin"

measurement_units = {
    1: ALBEDO,
    2: KELVIN,
    3: KELVIN,
    4: KELVIN,
    5: KELVIN,
}

# Nominal square IGFOV at nadir in km

square_igfov_at_nadir_in = {
    1: 1.0,
    2: 4.0,
    3: 8.0,
    4: 4.0,
    5: 4.0,
}

square_igfov_at_nadir_op = {
    1: 1.0,
    2: 4.0,
    3: 4.0,
    4: 4.0,
    5: 4.0,
}

square_igfov_at_nadir = {
    "G08": square_igfov_at_nadir_in,
    "G09": square_igfov_at_nadir_in,
    "G10": square_igfov_at_nadir_in,
    "G11": square_igfov_at_nadir_in,
    "G12": square_igfov_at_nadir_in,
    "G13": square_igfov_at_nadir_in,
    "G14": square_igfov_at_nadir_op,
    "G15": square_igfov_at_nadir_op,
}



"""
Calculate latitude and longitude grids data.

Notes
-----
See GOES-R Product User Guide (PUG) Volume 5 (L2 products) Section 4.2.8
for details & example of calculations.

Functions
---------
calculate_latlon_grid_goesdl
    Calculate latitude and longitude grids using an optimized version of
    classic's algorithm.
"""

from numpy import (
    arctan,
    cos,
    float32,
    float64,
    meshgrid,
    power,
    rad2deg,
    sin,
    sqrt,
)

from ..utils.array import ArrayFloat32, ArrayFloat64
from .helpers import make_common_mask
from .geostationary_parameters import GeostationaryParameters


def geos_to_latlon_grid_goesdl(
    projection_info: GeostationaryParameters,
) -> tuple[ArrayFloat32, ArrayFloat32]:
    """
    Calculate latitude and longitude grids.

    Calculate latitude and longitude from GOES ABI fixed grid projection
    data using an optimized version of NOAA's algorithm. GOES ABI fixed
    grid projection is a map projection relative to the GOES satellite.

    Units: latitude in °N (°S < 0), longitude in °E (°W < 0)

    Notes
    -----
    For information on GOES Imager Projection and GOES orbit geometry,
    see [1]_ and Section 4.2. of [3]_. For a Python demonstration on
    calculating latitude and longitude from GOES Imager Projection
    information, see [2]_. The code snippet in this class is based on
    the Python demonstration in [2]_.

    Parameters
    ----------
    projection_info : ProjectionParameters
        Object containing the satellite's projection information and
        GOES ABI fixed grid data.

    Returns
    -------
    tuple[ArrayFloat32, ArrayFloat32]
        A tuple containing the latitude and longitude grid data.

    References
    ----------
    .. [1] STAR Atmospheric Composition Product Training, "GOES Imager
        Projection (ABI Fixed Grid)", NOAA/NESDIS/STAR, 2024.
        https://www.star.nesdis.noaa.gov/atmospheric-composition-training/satellite_data_goes_imager_projection.php.
    .. [2] Aerosols and Atmospheric Composition Science Team, "Python
        Short Demo: Calculate Latitude and Longitude from GOES Imager
        Projection (ABI Fixed Grid) Information", NOAA/NESDIS/STAR,
        2024.
        https://www.star.nesdis.noaa.gov/atmospheric-composition-training/python_abi_lat_lon.php
    .. [3] GOES-R, " GOES-R Series Product Definition and User’s Guide
        (PUG), Volume 5: Level 2+ Products", Version 2.4,
        NASA/NOAA/NESDIS, 2022.
        https://www.ospo.noaa.gov/Organization/Documents/PUG/GS%20Series%20416-R-PUG-L2%20Plus-0349%20Vol%205%20v2.4.pdf
    """
    # Reorganize operations to leverage NumPy vectorization,
    # reducing redundant computations. This yields ~6x performance
    # improvement over the baseline implementation from [2].
    lambda_0 = projection_info.orbit.longitude_of_projection_origin

    r_orb = projection_info.orbital_radius
    r_eq = projection_info.globe.semi_major_axis
    r_pol = projection_info.globe.semi_minor_axis

    x_r = projection_info.x
    y_r = projection_info.y

    sin_x: ArrayFloat64 = sin(x_r)
    cos_x: ArrayFloat64 = cos(x_r)
    sin_y: ArrayFloat64 = sin(y_r)
    cos_y: ArrayFloat64 = cos(y_r)

    sin_x, sin_y = meshgrid(sin_x, sin_y)
    cos_x, cos_y = meshgrid(cos_x, cos_y)

    lat, lon = _transform_grid(
        (r_orb, r_eq, r_pol),
        (sin_x, sin_y),
        (cos_x, cos_y),
    )

    return lat.astype(float32), (lon + lambda_0).astype(float32)


def _transform_grid(
    params: tuple[float64, float64, float64],
    sin_xy: tuple[ArrayFloat64, ArrayFloat64],
    cos_xy: tuple[ArrayFloat64, ArrayFloat64],
) -> tuple[ArrayFloat64, ArrayFloat64]:
    """
    Calculate latitude and longitude grids.

    Helper function for functions calculate_latlon_grid_opti and
    calculate_latlon_grid_fast.

    Parameters
    ----------
    params : tuple[float64, float64, float64]
        Tuple containing the orbital radius, semi-major axis, and
        semi-minor axis of the Earth.
    sin_xy : tuple[ArrayFloat64, ArrayFloat64]
        Tuple containing the sine values of the x and y grid points.
    cos_xy : tuple[ArrayFloat64, ArrayFloat64]
        Tuple containing the cosine values of the x and y grid points.

    Returns
    -------
    tuple[ArrayFloat32, ArrayFloat32]
        A tuple containing the latitude and longitude grid data.

    Notes
    -----
    Based on NOAA/NESDIS/STAR Aerosols and Atmospheric Composition
    Science Team's code found on [2], which is based on the GOES-R
    Product User Guide (PUG) Volume 5 (L2 products) Section 4.2.8.
    Retrieved at 2025-02-24.
    """

    r_orb, r_eq, r_pol = params
    sin_x, sin_y = sin_xy
    cos_x, cos_y = cos_xy

    # Equations to calculate latitude and longitude
    a_var = power(sin_x, 2.0) + (
        power(cos_x, 2.0)
        * (
            power(cos_y, 2.0)
            + (((r_eq * r_eq) / (r_pol * r_pol)) * power(sin_y, 2.0))
        )
    )

    b_var = -2.0 * r_orb * cos_x * cos_y
    c_var = (r_orb**2.0) - (r_eq**2.0)
    r_s = (-1.0 * b_var - sqrt((b_var**2) - (4.0 * a_var * c_var))) / (
        2.0 * a_var
    )

    s_x = r_s * cos_x * cos_y
    s_y = -r_s * sin_x
    s_z = r_s * cos_x * sin_y

    abi_lat: ArrayFloat64 = rad2deg(
        arctan(
            ((r_eq * r_eq) / (r_pol * r_pol))
            * (s_z / sqrt(((r_orb - s_x) * (r_orb - s_x)) + (s_y * s_y)))
        )
    )
    abi_lon: ArrayFloat64 = rad2deg(arctan(s_y / (s_x - r_orb)))

    abi_lat, abi_lon = make_common_mask(abi_lat, abi_lon)

    return abi_lat, abi_lon



import unittest

from goesdl.utils.url import URL as URL


class TestUrlMethods(unittest.TestCase):
    HOST_URL = "http://example.com"
    BASE_URL = "http://example.com/path/"
    FOLDER_URL = "http://example.com/path/to"
    FILE_URL = "http://example.com/path/to/file"
    FULL_URL = (
        "http://user:pass@example.com:80/"
        "path/to/file;parameters?query#fragment"
    )

    def test_abspath_current_dir(self) -> None:
        original_url = "http://example.com/path/./to/file"
        expected_url = self.FILE_URL
        self.assertEqual(URL.abspath(original_url), expected_url)

    def test_abspath_sub_dir(self) -> None:
        original_url = "http://example.com/path/../to/file"
        expected_url = "http://example.com/to/file"
        self.assertEqual(URL.abspath(original_url), expected_url)

    def test_join_single_part_override(self) -> None:
        original_url = self.BASE_URL
        override_url = self.FILE_URL
        self.helper_join_single_part(original_url, override_url, override_url)

    def test_join_single_part_override_alt(self) -> None:
        original_url = self.FILE_URL
        override_url = self.FOLDER_URL
        self.helper_join_single_part(original_url, override_url, override_url)

    def test_join_single_part_override_same(self) -> None:
        override_url = self.FILE_URL
        self.helper_join_single_part(override_url, override_url, override_url)

    def test_join_multiple_parts_to_base(self) -> None:
        original_url = self.BASE_URL
        expected_url = self.FILE_URL
        self.helper_join_multiple_parts(
            original_url, "to/", "file", expected_url
        )

    def test_join_multiple_parts_host(self) -> None:
        original_url = self.HOST_URL
        expected_url = self.FILE_URL
        self.helper_join_multiple_parts(
            original_url, "path/to/", "file", expected_url
        )

    def test_join_single_part_to_base(self) -> None:
        original_url = self.BASE_URL
        expected_url = self.FILE_URL
        self.helper_join_single_part(original_url, "to/file", expected_url)

    def test_join_single_part_to_host(self) -> None:
        original_url = self.HOST_URL
        expected_url = self.FILE_URL
        self.helper_join_single_part(
            original_url, "path/to/file", expected_url
        )

    def test_parse_scheme(self) -> None:
        parsed_url = URL.parse(self.FULL_URL)
        self.assertEqual(parsed_url.scheme, "http")

    def test_parse_location(self) -> None:
        parsed_url = URL.parse(self.FULL_URL)
        self.assertEqual(parsed_url.netloc, "user:pass@example.com:80")

    def test_parse_path(self) -> None:
        parsed_url = URL.parse(self.FULL_URL)
        self.assertEqual(parsed_url.path, "/path/to/file")

    def test_parse_path_parameters(self) -> None:
        parsed_url = URL.parse(self.FULL_URL)
        self.assertEqual(parsed_url.params, "parameters")

    def test_parse_path_query(self) -> None:
        parsed_url = URL.parse(self.FULL_URL)
        self.assertEqual(parsed_url.query, "query")

    def test_parse_path_fragment(self) -> None:
        parsed_url = URL.parse(self.FULL_URL)
        self.assertEqual(parsed_url.fragment, "fragment")

    def test_parse_path_username(self) -> None:
        parsed_url = URL.parse(self.FULL_URL)
        self.assertEqual(parsed_url.username, "user")

    def test_parse_path_password(self) -> None:
        parsed_url = URL.parse(self.FULL_URL)
        self.assertEqual(parsed_url.password, "pass")

    def test_parse_path_hostname(self) -> None:
        parsed_url = URL.parse(self.FULL_URL)
        self.assertEqual(parsed_url.hostname, "example.com")

    def test_parse_path_port(self) -> None:
        parsed_url = URL.parse(self.FULL_URL)
        self.assertEqual(parsed_url.port, 80)

    def helper_join_multiple_parts(
        self,
        base_url: str,
        part_url_1: str,
        part_url_2: str,
        expected_url: str,
    ) -> None:
        joined_url = URL.join(base_url, part_url_1, part_url_2)
        self.assertEqual(joined_url, expected_url)

    def helper_join_single_part(
        self, base_url: str, part_url: str, expected_url: str
    ) -> None:
        joined_url = URL.join(base_url, part_url)
        self.assertEqual(joined_url, expected_url)



"""
Calculate latitude and longitude grids data.

Notes
-----
See GOES-R Product User Guide (PUG) Volume 5 (L2 products) Section 4.2.8
for details & example of calculations.

Functions
---------
calculate_latlon_grid
    Calculate latitude and longitude grids from geostationary grid.
"""

from typing import Literal

from ..utils.array import ArrayFloat32
from .geos_to_latlon_grid_cartopy import geos_to_latlon_grid_cartopy
from .geos_to_latlon_grid_goesdl import geos_to_latlon_grid_goesdl
from .geos_to_latlon_grid_pyproj import geos_to_latlon_grid_pyproj
from .geostationary_parameters import GeostationaryParameters


def geos_to_latlon_grid(
    parameters: GeostationaryParameters,
    algorithm: Literal["cartopy", "goesdl", "pyproj"] = "goesdl",
) -> tuple[ArrayFloat32, ArrayFloat32]:
    if algorithm == "cartopy":
        return geos_to_latlon_grid_cartopy(parameters)
    if algorithm == "pyproj":
        return geos_to_latlon_grid_pyproj(parameters)
    return geos_to_latlon_grid_goesdl(parameters)



"""
GOES dataset metadata extraction.

This module provides classes to extract and represent dataset metadata
from GOES satellite netCDF data files.

Classes
-------
GOESDatasetInfo
    Represent GOES dataset metadata attributes.
"""

from ..netcdf import DatasetView, dimension


class GOESDatasetInfo(DatasetView):
    """
    Hold GOES dataset metadata information.

    Attributes
    ----------
    institution : str
        The institution responsible for the dataset.
    project : str
        The project under which the dataset was created.
    production_site : str
        The site where the dataset was produced.
    production_environment : str
        The environment where the dataset was produced.
    spatial_resolution : str
        The spatial resolution of the dataset.
    orbital_slot : str
        The orbital slot of the satellite.
    platform_ID : str
        The platform identifier of the satellite.
    instrument_type : str
        The type of instrument used to collect the data.
    scene_id : str
        The identifier for the scene.
    instrument_ID : str
        The identifier for the instrument.
    dataset_name : str
        The name of the dataset.
    title : str
        The title of the dataset.
    summary : str
        A summary of the dataset.
    keywords : str
        Keywords associated with the dataset.
    keywords_vocabulary : str
        The vocabulary used for the keywords.
    license : str
        The license under which the dataset is released.
    processing_level : str
        The processing level of the dataset.
    cdm_data_type : str
        The CDM data type of the dataset.
    date_created : str
        The date the dataset was created.
    time_coverage_start : str
        The start time of the data coverage.
    time_coverage_end : str
        The end time of the data coverage.
    timeline_id : str
        The identifier for the timeline.
    production_data_source : str
        The data source used to produce the dataset.
    y : int
        The number of rows in the dataset.
    x : int
        The number of columns in the dataset.
    """

    institution: str
    project: str
    production_site: str
    production_environment: str
    spatial_resolution: str
    orbital_slot: str
    platform_ID: str  # NOSONAR
    instrument_type: str
    scene_id: str
    instrument_ID: str  # NOSONAR
    dataset_name: str
    title: str
    summary: str
    keywords: str
    keywords_vocabulary: str
    license: str
    processing_level: str
    cdm_data_type: str
    date_created: str
    time_coverage_start: str
    time_coverage_end: str
    timeline_id: str
    production_data_source: str
    y: int = dimension.size()
    x: int = dimension.size()



"""
Provide functionality for accessing GOES-R Series and GridSat datasets.

This package provides classes to download and extract information from
GOES-R Series and GridSat netCDF data files. It uses the `requests`
and `boto3` packages to download the files and the `netCDF4` package to
read the data files.
"""

from .info import __package_id__, __package_name__, __version__

__all__ = [
    "__package_id__",
    "__package_name__",
    "__version__",
]



"""
Provide locator for GridSat family of imagery datasets products.

Classes:
    - GridSatProductLocator: Abstract a product locator for GridSat
      family of imagery datasets.
"""

from abc import abstractmethod
from dataclasses import dataclass
from datetime import datetime

from ..dataset import ProductLocatorGG
from .constants import GRIDSAT_FILE_SUFFIX


@dataclass(eq=False, frozen=True)
class GridSatProductLocator(ProductLocatorGG):
    """
    Abstract a product locator for GridSat family of imagery datasets.

    This abstract class implements the `ProductLocator` interface for
    a generic GridSat dataset product locator by inheriting from the
    `ProductLocatorGG` abstract class. The class implements the common
    specifications and naming conventions for products in the GridSat
    datasets, and serves as a base class for more specialised product
    locator classes. It also defines the common attributes and abstract
    methods for all GridSat dataset product locators.

    Instances of this class are responsible for generating a list of
    folder paths based on the dataset's directory structure and naming
    conventions, product details, and the specified date range. The
    generated paths cover each year or month interval within the
    required period according on the dataset's temporal granularity;
    paths to the folders containing the initial and final dates are
    included in the list.

    `GridSatProductLocator` objects are responsible for verifying if a
    given filename matches the product filename pattern based on the
    dataset's naming conventions and product specifications invoking the
    `match()` method, and for extracting the corresponding `datetime`
    information from the product's filename via the `get_datetime()`
    method.

    Also, the class provides the implementation for the abstract helper
    methods defined in the `ProductLocatorGG` abstract class. The
    `GridSatProductLocator` class is the workhorse for all GridSat
    dataset product locators.

    Notes
    -----
    Subclasses are responsible for initialising the attributes with the
    appropriate values for the dataset and product details, and provide
    implementation for the `get_base_url()` method declared by the
    `ProductLocator` interface. Definition for abstract methods declared
    by this class: `next_time()` and `normalize_times()` should also be
    provided. Refer to their individual documentation for details.

    Notes
    -----
    Currently, the 'Geostationary IR Channel Brightness Temperature
    - GridSat B1' and the 'Geostationary Operational Environmental
    Satellites - GOES/CONUS' datasets's products are supported.

    Caution
    -------
    Members of this class not defined by the `ProductLocator` interface
    are helper methods and can be considered as implementation details,
    even though they are defined as part of the public API. In future
    releases, these methods may be moved to a private scope, suffer name
    changes, or be removed altogether.

    Methods
    -------
    get_date_format()
        Return the date format specification for the GridSat product's
        filename.
    get_paths(datetime_ini: datetime, datetime_fin: datetime)
        Generate a list of paths containing the product files for the
        specified date range.
    get_prefix()
        Generate the prefix for the GridSat product's filename.
    get_suffix()
        Generate the suffix for the GridSat product's filename.
    get_timestamp_pattern()
        Return the timestamp pattern for the GridSat product's filename.
    next_time(current_time: datetime)
        Get the next time interval.
    normalize_times(datetime_ini: datetime, datetime_fin: datetime)
        Normalise the initial and final datetimes.

    Attributes
    ----------
    name : str
        The name of the GridSat dataset product. A dataset can have
        multiple products. E.g. "B1", "GOES". Due to how the GridSat
        dataset is structured, the name is always a single string.
    origins : list[str]
        The list of origins of the GridSat product, namely one or more
        satellite identifier, e.g. "goes08". Multi-origin datasets, like
        GridSat-B1, set this attribute to an empty list.
    versions : list[str]
        The version or list of versions of the GridSat product; e.g.
        "v01r01".
    file_date_format : str
        The specification of the date format used in the GridSat product
        filename.
    file_date_pattern : str
        The regex pattern of the date format used in the GridSat product
        filename.
    file_prefix : str
        The prefix for the GridSat product's filenames.
    path_date_format : str
        The specification of the path's date format convention used in
        the dataset.
    path_prefix : str
        The path prefix to the dataset directories.
    """  # noqa: E501

    name: str
    origins: list[str]
    versions: list[str]
    file_date_format: str
    file_date_pattern: str
    file_prefix: str
    path_date_format: str
    path_prefix: str

    def get_date_format(self) -> str:
        """
        Return the date format specification for the product's filename.

        Generates and returns the date format specification for
        the product's filename based on the GridSat dataset product
        filename's date and time format conventions. The date format
        specification string is used to parse the product's filename
        and extract the `datetime` information.

        Returns
        -------
        str
            The date format specification for the GridSat product's
            filename.
        """
        return self.file_date_format

    def get_paths(
        self, datetime_ini: datetime, datetime_fin: datetime
    ) -> list[str]:
        """
        Generate a list of dataset directory paths.

        This method generates a list of directory paths within the
        dataset based on the folder structure and naming conventions,
        temporal granularity, and the specified date range. Paths to the
        folders containing the initial and final dates are included in
        the list.

        Parameters
        ----------
        datetime_ini : datetime
            The initial datetime for the desired data.
        datetime_fin : datetime
            The final datetime for the desired data.

        Returns
        -------
        list[str]
            A list of strings representing the paths to dataset
            directories containing the product files for the
            specified date range.
        """
        current_time: datetime
        end_time: datetime
        current_time, end_time = self.normalize_times(
            datetime_ini, datetime_fin
        )

        folder_paths: list[str] = []

        while current_time <= end_time:
            folder_path: str = current_time.strftime(self.path_date_format)
            folder_paths.append(f"{self.path_prefix}{folder_path}/")

            current_time = self.next_time(current_time)

        return folder_paths

    def get_prefix(self) -> str:
        """
        Return the prefix for the product's filename.

        Generates and returns the prefix for the GridSat product's
        filename based on product-specific information like dataset
        and product's name, instrument and origin's identifier, etc.

        Returns
        -------
        str
            The prefix for the GridSat product's filename.
        """
        sorted_origins: list[str] = sorted(self.origins)
        origins: str = (
            rf"\.(?:{'|'.join(sorted_origins)})" if self.origins else ""
        )

        return rf"{self.file_prefix}-{self.name}{origins}\."

    def get_suffix(self) -> str:
        """
        Return the suffix for the product's filename.

        Generates and returns the suffix for the product's filename
        based on product-specific information like product's version
        origin's identifier, and file suffix (extension).

        Returns
        -------
        str
            The suffix for the GridSat product's filename.
        """
        sorted_versions: list[str] = sorted(self.versions)

        return rf"\.(?:{'|'.join(sorted_versions)})\{GRIDSAT_FILE_SUFFIX}"

    def get_timestamp_pattern(self) -> str:
        """
        Return the timestamp regex pattern for the product's filename.

        Generates and returns the timestamp regular expression
        pattern for the product's filename based on the dataset
        product filename's date and time format conventions. The
        timestamp regex pattern is used to extract the substring
        containing the timestamp from the product's filename
        before extracting the `datetime` information.

        Returns
        -------
        str
            The timestamp regex pattern for the GridSat product's
            filename.
        """
        return f"({self.file_date_pattern})"

    @abstractmethod
    def next_time(self, current_time: datetime) -> datetime:
        """
        Get the next time interval.

        Get the next time interval based on the current time interval.

        Parameters
        ----------
        current_time : datetime
            The current time interval.

        Returns
        -------
        datetime
            The next time interval.
        """

    @abstractmethod
    def normalize_times(
        self, datetime_ini: datetime, datetime_fin: datetime
    ) -> tuple[datetime, datetime]:
        """
        Normalise the initial and final datetimes.

        Normalise the initial and final datetimes to the nearest
        time interval based on the dataset. The initial datetime is
        normalised to the start of the time interval and the final
        datetime is normalised to the end of the time interval.

        Parameters
        ----------
        datetime_ini : datetime
            The initial datetime.
        datetime_fin : datetime
            The final datetime.

        Returns
        -------
        tuple[datetime, datetime]
            A tuple containing the normalised initial and final
            datetimes.
        """



import math

import gudhi as gd
import matplotlib.pyplot as plt
import numpy as np
from scipy import ndimage


def remove_small_blobs(binary_image, size):
    labeled, nobject = ndimage.label(1 - binary_image)
    for label in range(0, nobject):
        masked = labeled == label
        area = np.count_nonzero(masked)
        if area <= size:
            binary_image[masked] = 1

    return binary_image


def create_complex(field_map, verbose=False):
    # Crear el complejo cúbico
    cubical_complex = gd.CubicalComplex(top_dimensional_cells=field_map)

    if verbose:
        print(
            f"Cubical complex is of dimension {cubical_complex.dimension()} - {cubical_complex.num_simplices()} cubes."
        )

    return cubical_complex


def maximum_persistence(n, persistence_pairs):
    max_birth = 0
    max_death = 0
    max_lifetime = 0
    for h, (birth, death) in persistence_pairs:
        if h == n:
            lifetime = death - birth
            if lifetime > max_lifetime and lifetime != math.inf:
                max_lifetime = lifetime
                max_birth = birth
                max_death = death

    return max_lifetime, (max_birth, max_death)


def plot_persistence_diagram(persistence_pairs):
    # Extract birth and death times
    labels = [pair[0] for pair in persistence_pairs]
    births = [pair[1][0] for pair in persistence_pairs]
    deaths = [pair[1][1] for pair in persistence_pairs]
    deaths = [255 if value > 255 else value for value in deaths]

    # # Plot the persistence diagram
    plt.figure(figsize=(8, 6))
    plt.scatter(births, deaths, c=labels)
    plt.xlabel("Birth")
    plt.ylabel("Death")
    plt.title("Persistence Diagram")
    plt.show()


def plot_persistence_diagram(persistence_pairs):
    # Graficar el diagrama de persistencia
    gd.plot_persistence_diagram(persistence_pairs)
    plt.xlabel("Birth")
    plt.ylabel("Death")
    plt.title("Persistence Diagram")

    plt.tight_layout()
    plt.show()



"""
Provide the DatasourceHTTP class for handling HTTP-based data sources.

Classes:
    DatasourceHTTP: Handle HTTP-based data sources.
"""

import re
import socket
from urllib.parse import ParseResult

import requests

from ..dataset import ProductLocator
from ..utils.headers import APPLICATION_NETCDF4, TEXT_HTML, RequestHeaders
from ..utils.url import URL
from .datasource_base import DatasourceBase
from .datasource_cache import DatasourceCache

HTTP_STATUS_OK = 200


class DatasourceHTTP(DatasourceBase):
    """
    Handle HTTP-based data sources.

    Provide methods to interact with HTTP folders and files, either
    through a base URL or a `ProductLocator` object.

    Methods
    -------
    download_file(file_path: str)
        Retrieve a file from the datasource and save it into the local
        repository.
    list_files(dir_path: str)
        List the contents of a remote directory.
    """

    def __init__(
        self,
        locator: str | ProductLocator,
        cache: float | DatasourceCache | None = None,
    ) -> None:
        """
        Initialize the DatasourceHTTP object.

        Parameters
        ----------
        locator : str | ProductLocator
            The base URL of a HTTP-based data sources or a
            `ProductLocator` object.
        cache : float | DatasourceCache | None, optional
            The cache expiration time in seconds, by default None.

        Raises
        ------
        ValueError
            If the resource does not exist or the user has no access.
        """
        base_url: str = (
            locator
            if isinstance(locator, str)
            else locator.get_base_url("HTTP")[0]
        )

        url_parts: ParseResult = URL.parse(base_url)

        host_name: str = url_parts.netloc
        base_path = url_parts.path

        if not self._host_exists(host_name):
            raise ValueError(
                f"Host '{host_name}' does not exist or is out of service."
            )
        if not self._path_exists(base_url):
            raise ValueError(
                f"Path '{base_path}' does not exist or you have no access."
            )

        super().__init__(base_url, cache)

    def download_file(self, file_path: str) -> bytes:
        """
        Download a file from the datasource into the local repository.

        Get a file from a remote location or local repository. The path
        provided must be relative to the base URL and local repository
        root directory. The remote path is reconstructed in the local
        repository.

        Parameters
        ----------
        file_path : str
            The path to the remote file to be downloaded.

        Returns
        -------
        bytes
            The content of the file as a byte string.

        Raises
        ------
        RuntimeError
            If the file cannot be retrieved.
        """
        try:
            return self._retrieve_file(file_path)

        except requests.HTTPError as exc:
            raise RuntimeError(
                f"Unable to retrieve the file '{file_path}': {exc}"
            ) from exc

    def list_files(self, dir_path: str) -> list[str]:
        """
        List the contents of a directory.

        List the contents of a directory in a remote location. The path
        is relative to the base URL.

        Parameters
        ----------
        dir_path : str
            The path to the directory. The path is relative to the base
            URL.

        Returns
        -------
        list[str]
            A list of file names in the directory.
        """
        cached_links = self.cache.get_item(dir_path)

        if cached_links is not None:
            return cached_links

        folder_url = URL.join(self.base_url, dir_path)
        index_html = self._get_content(folder_url)

        if not index_html:
            return []

        href_links = re.findall(r'<a\s+href="([^"]+)"', index_html)
        href_links = [URL.join(folder_url, href) for href in href_links]
        href_links = [href.replace(self.base_url, "") for href in href_links]

        self.cache.add_item(dir_path, href_links)

        return href_links

    @staticmethod
    def _host_exists(host_name: str) -> bool:
        try:
            socket.gethostbyname(host_name)

        except socket.gaierror:  # as exc
            # Host does not exist or is out of service (log this!)
            return False

        return True

    @staticmethod
    def _path_exists(folder_url: str) -> bool:
        response = requests.head(folder_url, timeout=10)

        return response.status_code == HTTP_STATUS_OK

    @staticmethod
    def _get_content(folder_url: str) -> str:
        headers = RequestHeaders(accept=TEXT_HTML).headers

        response = requests.get(folder_url, headers=headers, timeout=15)

        if response.status_code == HTTP_STATUS_OK:
            response.encoding = response.apparent_encoding
            return response.text

        return ""

    def _retrieve_file(self, file_path: str) -> bytes:
        file_url = URL.join(self.base_url, file_path)

        headers = RequestHeaders(accept=APPLICATION_NETCDF4).headers

        response = requests.get(file_url, headers=headers, timeout=15)

        response.raise_for_status()

        if response.status_code != HTTP_STATUS_OK:
            raise requests.HTTPError("Request failure", response=response)

        return bytes(response.content)



"""
Provide the EnhacementPalette class for handling color enhancement
palettes.
"""

from pathlib import Path

from .constants import UNNAMED_TABLE
from .cpt_table import cpt_utility
from .eu_table import eu_utility
from .shared import ColorTable, DomainData, PaletteData, PaletteItem


class EnhacementPalette:
    """
    Represent a color enhancement color palette.

    This class provides methods to load, parse, and process McIDAS and
    GMT enhancement color tables, creating color dictionaries for
    Matplotlib colormaps.

    Attributes
    ----------
    name : str
        The name of the enhancement color palette.
    extent : DomainData
        The palette extent containing the minimum and maximum defined
        values.
    table : PaletteData
        The palette data containing color segments.
    stock : ColorStock
        The stock color values.

    Methods
    -------
    create_file(path, name, table, extent, rgb)
        Create a file with the enhancement color palette.
    from_file(path)
        Load a McIDAS or GMT enhancement color palette specification and
        create an EnhacementPalette instance.
    reverse()
        Reverse the enhancement color palette.
    save_to_file(path, name, rgb)
        Save the enhancement color palette.
    """

    name: str
    extent: DomainData
    table: PaletteData
    stock: ColorTable

    def __init__(
        self, name: str, specs: PaletteItem, extent: DomainData
    ) -> None:
        self.name = name
        self.extent = extent
        self.table = specs[0]
        self.stock = specs[1]

    @classmethod
    def from_file(cls, path: str | Path) -> "EnhacementPalette":
        """
        Load a McIDAS or GMT enhancement color table specification.

        Parse a McIDAS enhancement utility table (EU TABLE) or GMT color
        palette table (CPT TABLE) text file and create color dictionary
        for a Matplotlib colormap.

        Parameters
        ----------
        path : str or Path
            Path to the color table specification text file.

        Returns
        -------
        ColorSpec
            Tuple of color table and stock color values.

        Notes
        -----
        The Man computer Interactive Data Access System (McIDAS) is a
        research quality suite of applications used for decoding,
        analyzing, and displaying meteorological data developed by the
        University of Wisconsin-Madison Space Science and Engineering
        Center (UWisc/SSEC).

        - https://www.ssec.wisc.edu/mcidas/
        - https://www.unidata.ucar.edu/software/mcidas/

        The Generic Mapping Tools (GMT) is an open-source collection of
        tools for manipulating geographic and Cartesian data sets and
        producing PostScript illustrations ranging from simple x-y plots
        through maps to complex 3D perspective views.

        - https://www.generic-mapping-tools.org/
        """
        with open(path, "r", encoding="utf-8") as file:
            lines = file.readlines()

        items, name, extent = cls._parse_table(lines)

        return cls(name, items, extent)

    def save_to_file(
        self, path: str | Path, name: str = "", rgb: bool = False
    ) -> None:
        """
        Save the color table.

        Save the color table to a McIDAS enhancement utility table (EU
        TABLE) or GMT color palette table (CPT TABLE) text file.

        Parameters
        ----------
        path : str or Path
            Path to the file where the color table will be saved.
        name : str
            Name of the color table. Override the actual name if provided.
        rgb : bool, optional
            Flag indicating if the color model is RGB, by default False.
        """
        if not name and self.name != UNNAMED_TABLE:
            name = self.name

        eu_utility.create_file(path, name, self.table, self.extent, rgb)

    @staticmethod
    def _parse_table(lines: list[str]) -> tuple[PaletteItem, str, DomainData]:
        # Try parse a EU file first (if EU file detected)
        if eu_utility.is_eu_table(lines[0]):
            return eu_utility.parse_eu_table(lines)

        # Try parse a CPT file
        return cpt_utility.parse_cpt_table(lines)



"""
Define and export constants used across the GOES-DL downloader module.

These constants include date and time formats, and other fixed values
that are used throughout the module.

Classes
-------
DownloadStatus:
    Enumeration for download status.

Attributes
----------
ISO_TIMESTAMP_FORMAT : str
    The format string for ISO 8601 timestamps.
TIME_TOLERANCE_DEFAULT : int
    The default value for the time tolerance.
TIME_TOLERANCE_MAX : int
    The upper limit for the time tolerance.
TIME_TOLERANCE_MIN : int
    The lower limit for the time tolerance.
TIME_TOLERANCE_SINGLE : int
    The time tolerance value for sigle file downloading.
"""

from enum import Enum

ISO_TIMESTAMP_FORMAT: str = "%Y-%m-%dT%H:%M:%S%z"

# Default values for the time tolerance (in seconds).
TIME_TOLERANCE_DEFAULT: int = 60
TIME_TOLERANCE_MAX: int = 300
TIME_TOLERANCE_MIN: int = 30
TIME_TOLERANCE_SINGLE: int = 300


class DownloadStatus(Enum):
    """
    Enumeration for download status.

    Attributes
    ----------
    SUCCESS
        Indicates that the download was successful.
    FAILED
        Indicates that the download failed.
    ALREADY
        Indicates that the file was already downloaded.
    """

    SUCCESS = 0
    FAILED = 1
    ALREADY = 2



from collections.abc import Callable
from typing import Any, cast

from cartopy.crs import Globe, PlateCarree, Projection
from netCDF4 import Dataset  # pylint: disable=no-name-in-module
from numpy import concatenate, flatnonzero, float32, meshgrid

from ..geodesy import RectangularRegion
from ..netcdf import DatasetView, HasStrHelp, attribute, variable
from ..utils.array import ArrayFloat32
from .metadata import CoordinateMetadata, VariableMetadata

LimitType = tuple[int, int]
MetadataType = dict[str, CoordinateMetadata | VariableMetadata]


class GSLatLonData(HasStrHelp):
    lon: ArrayFloat32
    lat: ArrayFloat32


class GeodeticSummary(DatasetView):

    geospatial_lat_min: float32 = attribute()
    geospatial_lat_max: float32 = attribute()
    geospatial_lat_units: str = attribute()
    geospatial_lat_resolution: float32 = attribute()
    geospatial_lon_min: float32 = attribute()
    geospatial_lon_max: float32 = attribute()
    geospatial_lon_units: str = attribute()
    geospatial_lon_resolution: float32 = attribute()


class GSLatLonGrid(GSLatLonData):

    region: RectangularRegion

    lon: ArrayFloat32
    lat: ArrayFloat32

    lon_limits: LimitType
    lat_limits: LimitType

    metadata: MetadataType

    crs: Projection

    summary: GeodeticSummary

    def __init__(
        self,
        record: Dataset,
        region: RectangularRegion | None = None,
        delta: int = 5,
        corners: bool = False,
    ) -> None:
        # Validate delta parameter (subsampling increment step)
        if not (1 <= delta <= 10):
            raise ValueError(
                "'delta' must be an integer between 1 and 10, inclusive"
            )

        # Extract the region of interest...
        if region:
            data, lon_limits, lat_limits = self._slice(
                record, region, delta, corners
            )

        # ...or extract the entire field of view
        else:
            region = self._extract_region(record)
            data, lon_limits, lat_limits = self._full_frame(record, corners)

        self.region = region

        self.lon, self.lat = meshgrid(data.lon, data.lat)

        self.lon_limits = lon_limits
        self.lat_limits = lat_limits

        # Create the source projection (Platé-Carrée projection on GRS80
        # ellipsoid)

        source_globe = Globe(ellipse="GRS80")

        self.crs = PlateCarree(central_longitude=0.0, globe=source_globe)

        self.metadata = self._get_metadata(record)

        self.summary = GeodeticSummary(record)

    @staticmethod
    def _extract(
        record: Dataset,
        step: int | None,
        lon_limits: LimitType | None,
        lat_limits: LimitType | None,
    ) -> "GSLatLonData":
        def subsample(limits: LimitType | None) -> Callable[[Any], Any]:
            def closure(x: Any) -> Any:
                begin, end = limits or (None, None)
                skip = step or None
                return x[begin:end:skip]

            return closure

        class _LatLonData(DatasetView):
            lon: ArrayFloat32 = variable("lon").data(
                filter=subsample(lon_limits)
            )
            lat: ArrayFloat32 = variable("lat").data(
                filter=subsample(lat_limits)
            )

        data = _LatLonData(record)

        return cast(GSLatLonData, data)

    @staticmethod
    def _extract_bounds(
        record: Dataset,
        lon_limits: LimitType | None,
        lat_limits: LimitType | None,
    ) -> "GSLatLonData":
        def subsample(limits: LimitType | None) -> Callable[[Any], Any]:
            def closure(x: Any) -> Any:
                begin, end = limits or (None, None)
                coord = x[begin:end]
                return concatenate((coord[:, 0], coord[-1:, -1]))

            return closure

        class _LatLonData(DatasetView):
            lon: ArrayFloat32 = variable("lon_bounds").data(
                filter=subsample(lon_limits)
            )
            lat: ArrayFloat32 = variable("lat_bounds").data(
                filter=subsample(lat_limits)
            )

        data = _LatLonData(record)

        return cast(GSLatLonData, data)

    @staticmethod
    def _extract_bounds_metadata(
        record: Dataset, name: str
    ) -> VariableMetadata:
        coordinate = variable(name)

        class _LatLonMetata(DatasetView):
            long_name: str = coordinate.attribute()
            comment: str = coordinate.attribute()
            units: str = coordinate.attribute()
            shape: tuple[int] = coordinate.attribute()

        metadata = _LatLonMetata(record)

        return VariableMetadata(metadata)

    @staticmethod
    def _extract_metadata(record: Dataset, name: str) -> CoordinateMetadata:
        coordinate = variable(name)

        class _LatLonMetata(DatasetView):
            long_name: str = coordinate.attribute()
            standard_name: str = coordinate.attribute()
            units: str = coordinate.attribute()
            content_type: str = coordinate.attribute("coverage_content_type")
            axis: str = coordinate.attribute()
            shape: tuple[int] = coordinate.attribute()

        metadata = _LatLonMetata(record)

        return CoordinateMetadata(metadata)

    @staticmethod
    def _extract_region(record: Dataset) -> RectangularRegion:
        def subsample(x: Any) -> Any:
            skip = x.shape[0] - 1
            return x[::skip]

        class _LatLonData(DatasetView):
            lon: ArrayFloat32 = variable("lon").data(filter=subsample)
            lat: ArrayFloat32 = variable("lat").data(filter=subsample)

        data = _LatLonData(record)

        domain = (
            (float(data.lon[0]), float(data.lon[-1])),
            (float(data.lat[0]), float(data.lat[-1])),
        )

        return RectangularRegion(domain)

    @staticmethod
    def _find_limits(
        coord: ArrayFloat32,
        min_max: tuple[float, float],
        delta: int,
        offset: int = 0,
    ) -> LimitType:
        min_value, max_value = min_max

        min_indices = flatnonzero(coord < min_value)
        min_bound = min_indices[-1] if min_indices.size > 0 else 0

        max_indices = flatnonzero(coord > max_value)
        max_bound = max_indices[0] if max_indices.size > 0 else -1
        max_bound = max_bound + 1 if max_bound >= 0 else coord.size

        return int(min_bound * delta + offset), int(max_bound * delta + offset)

    @classmethod
    def _full_frame(
        cls,
        record: Dataset,
        corners: bool,
    ) -> tuple["GSLatLonData", LimitType, LimitType]:
        if corners:
            data = cls._extract_bounds(record, None, None)
            lon_limits = 0, data.lon.size - 1
            lat_limits = 0, data.lat.size - 1
        else:
            data = cls._extract(record, None, None, None)
            lon_limits = 0, data.lon.size
            lat_limits = 0, data.lat.size

        return data, lon_limits, lat_limits

    @classmethod
    def _get_metadata(cls, record: Dataset) -> MetadataType:
        return {
            "lon": cls._extract_metadata(record, "lon"),
            "lat": cls._extract_metadata(record, "lat"),
            "lon_bounds": cls._extract_bounds_metadata(record, "lon_bounds"),
            "lat_bounds": cls._extract_bounds_metadata(record, "lat_bounds"),
        }

    @classmethod
    def _slice(
        cls,
        record: Dataset,
        region: RectangularRegion,
        delta: int,
        corners: bool,
    ) -> tuple["GSLatLonData", LimitType, LimitType]:
        data = cls._extract(record, delta, None, None)

        lon_limits = cls._find_limits(data.lon, region.lon_bounds, delta)
        lat_limits = cls._find_limits(data.lat, region.lat_bounds, delta)

        if delta > 1:
            data = cls._extract(record, None, lon_limits, lat_limits)

            lon_limits = cls._find_limits(
                data.lon, region.lon_bounds, 1, lon_limits[0]
            )
            lat_limits = cls._find_limits(
                data.lat, region.lat_bounds, 1, lat_limits[0]
            )

        if corners:
            data = cls._extract_bounds(record, lon_limits, lat_limits)
        else:
            data = cls._extract(record, None, lon_limits, lat_limits)

        return data, lon_limits, lat_limits

    @property
    def globe(self) -> Globe:
        return self.crs.globe



"""
Export locators for GOES-R Series imagery dataset products.

Classes:
    - GOESProductLocatorABI: All primary and derived Advanced Baseline
      Imager (ABI) products.
    - GOESProductLocatorABIDC: All derived ABI products supporting
      channels.
    - GOESProductLocatorDMW: ABI Derived Motion Winds (DMW).
    - GOESProductLocatorDMWV: ABI Derived Motion WV Winds (DMWV).
    - GOESProductLocatorABIDP: All derived ABI products.
    - GOESProductLocatorGLM: All Geostationary Lightning Mapper (GLM)
      products.
    - GOESProductLocatorLCFA: GLM Lightning Cluster-Filter Algorithm
      (LCFA).
    - GOESProductLocatorABIPM: Multi-band primary ABI products.
    - GOESProductLocatorMCMIP: Multi-band CMIP Product.
    - GOESProductLocatorABIPP: All primary ABI products.
    - GOESProductLocatorCMIP: ABI Cloud and Moisture Imagery Product
      (CMIP).
    - GOESProductLocatorRad: ABI Radiance Product (Rad).
"""

from .locator_abi import GOESProductLocatorABI
from .locator_dc import (
    GOESProductLocatorABIDC,
    GOESProductLocatorDMW,
    GOESProductLocatorDMWV,
)
from .locator_dp import GOESProductLocatorABIDP
from .locator_glm import GOESProductLocatorGLM
from .locator_lp import GOESProductLocatorLCFA
from .locator_pm import GOESProductLocatorABIPM, GOESProductLocatorMCMIP
from .locator_pp import (
    GOESProductLocatorABIPP,
    GOESProductLocatorCMIP,
    GOESProductLocatorRad,
)

GOESDerivedProductLocator = GOESProductLocatorABIDP
GOESDerivedWithCannelProductLocator = GOESProductLocatorABIDC
GOESLightningMapperProductLocator = GOESProductLocatorGLM
GOESPrimaryMultibandProductLocator = GOESProductLocatorABIPM
GOESPrimaryProductLocator = GOESProductLocatorABIPP

__all__ = [
    "GOESProductLocatorABI",
    "GOESProductLocatorABIDC",
    "GOESProductLocatorDMW",
    "GOESProductLocatorDMWV",
    "GOESProductLocatorABIDP",
    "GOESProductLocatorGLM",
    "GOESProductLocatorLCFA",
    "GOESProductLocatorABIPM",
    "GOESProductLocatorMCMIP",
    "GOESProductLocatorABIPP",
    "GOESProductLocatorCMIP",
    "GOESProductLocatorRad",
    "GOESDerivedProductLocator",
    "GOESDerivedWithCannelProductLocator",
    "GOESLightningMapperProductLocator",
    "GOESPrimaryMultibandProductLocator",
    "GOESPrimaryProductLocator",
]



from typing import cast

from numpy import ceil, float64, linspace, pi

from ..utils.array import ArrayFloat64
from .aeqdproj_parameters import AeqdProjParameters
from .globe_parameters import GlobeParameters


class AzimuthalEquidistantParameters:
    """
    A class to store Azimuthal Equidistant projection parameters.

    Attributes
    ----------
    projection : AeqdProjectionParameters
        The parameters of the Azimuthal Equidistant projection.
    globe : GlobeParameters
        The parameters of the globe.
    x : ArrayFloat64
        The x-coordinate grid in meters.
    y : ArrayFloat64
        The y-coordinate grid in meters.
    """

    # Information about the projection
    projection: AeqdProjParameters

    # Information about the globe
    globe: GlobeParameters

    # Information about the fixed grid
    x: ArrayFloat64
    y: ArrayFloat64

    def __init__(
        self,
        projection: AeqdProjParameters,
        globe: GlobeParameters,
        xy_grid: tuple[ArrayFloat64, ArrayFloat64],
    ) -> None:
        """
        Initialize the class.

        Parameters
        ----------
        projection : AeqdProjectionParameters
            The parameters of the Azimuthal Equidistant projection.
        globe : GlobeParameters
            The parameters of the globe.
        xy_grid : tuple[ArrayFloat64, ArrayFloat64]
            The x and y grid data in meters.
        """
        self.projection = projection
        self.globe = globe
        self.x = xy_grid[0]
        self.y = xy_grid[1]

    @classmethod
    def create(
        cls,
        projection: AeqdProjParameters,
        globe: GlobeParameters,
        resolution: float64,
    ) -> "AzimuthalEquidistantParameters":
        horizontal_extent_m, vertical_extent_m = cls._get_size(
            projection, globe
        )

        width_px: float64 = ceil(horizontal_extent_m / resolution)
        height_px: float64 = ceil(vertical_extent_m / resolution)

        n_cols = int(width_px)
        n_rows = int(height_px)

        x_m = linspace(
            -horizontal_extent_m / 2.0,
            horizontal_extent_m / 2.0,
            n_cols,
            dtype=float64,
        )
        y_m = linspace(
            vertical_extent_m / 2.0,
            -vertical_extent_m / 2.0,
            n_rows,
            dtype=float64,
        )

        return AzimuthalEquidistantParameters(
            projection,
            globe,
            (x_m, y_m),
        )

    @classmethod
    def _get_size(
        cls, projection: AeqdProjParameters, globe: GlobeParameters
    ) -> tuple[float64, float64]:
        """
        Calculate the size of the projection.

        Parameters
        ----------
        projection : AeqdProjectionParameters
            The parameters of the Azimuthal Equidistant projection.
        resolution : float64
            The resolution in meters.

        Returns
        -------
        tuple[float64, float64]
            The width and height of the projection in kilometers.
        """
        width, height = projection.extent_size[:2]
        units = projection.extent_size[2]

        if units in {"arcsec", "arcmin", "deg"}:
            return cls._get_size_from_angle(
                width, height, units, globe.semi_major_axis
            )

        return cls._get_size_from_length(width, height, units)

    @staticmethod
    def _get_size_from_length(
        width: float64, height: float64, units: str
    ) -> tuple[float64, float64]:
        """
        Calculate the size of the projection from lengths.

        Parameters
        ----------
        width : float64
            The width of the projection in the specified unit of
            measurement.
        height : float64
            The height of the projection in the specified unit of
            measurement.
        units : str
            The unit of measurement for the width and height. It can be
            'm', or 'km'.

        Returns
        -------
        tuple[float64, float64]
            The width and height of the projection in meters.
        """
        if units == "km":
            width *= 1000.0
            height *= 1000.0

        return width, height

    @staticmethod
    def _get_size_from_angle(
        width: float64, height: float64, units: str, semi_major_axis: float64
    ) -> tuple[float64, float64]:
        """
        Calculate the size of the projection from angles.

        Parameters
        ----------
        width : float64
            The width of the projection in the specified unit of
            measurement.
        height : float64
            The height of the projection in the specified unit of
            measurement.
        units : str
            The unit of measurement for the width and height. It can be
            'arcsec', 'arcmin', or 'deg'.
        semi_major_axis : float64
            The semi-major axis of the globe in meters.

        Returns
        -------
        tuple[float64, float64]
            The width and height of the projection in meters.
        """
        deg_to_m = pi * semi_major_axis / 180.0

        if units == "arcsec":
            width /= 3600.0
            height /= 3600.0
        elif units == "arcmin":
            width /= 60.0
            height /= 60.0

        return width * deg_to_m, height * deg_to_m

    @property
    def extent(self) -> tuple[float64, float64, float64, float64]:
        """
        Calculate the extent of the projection.

        Returns
        -------
        tuple[float64, float64, float64, float64]
            The extent of the projection in kilometers.
            - The minimum x-coordinate.
            - The maximum x-coordinate.
            - The minimum y-coordinate.
            - The maximum y-coordinate.
        """
        return (
            self.x[0] / 1000.0,
            self.x[-2] / 1000.0,
            self.y[0] / 1000.0,
            self.y[-1] / 1000.0,
        )

    @property
    def x_km(self) -> ArrayFloat64:
        """
        Calculate the x-coordinate fixed grid in kilometers.

        Returns
        -------
        ArrayFloat64
            The x-coordinate fixed grid in kilometers.
        """
        return cast(ArrayFloat64, self.x / 1000.0)

    @property
    def y_km(self) -> ArrayFloat64:
        """
        Calculate the y-coordinate fixed grid in kilometers.

        Returns
        -------
        ArrayFloat64
            The y-coordinate fixed grid in kilometers.
        """
        return cast(ArrayFloat64, self.y / 1000.0)



from math import degrees, floor
from typing import Literal, cast

from numpy import arange, float32

from ..utils.array import ArrayFloat32

CenterType = tuple[float, float]
CoordLimits = tuple[float, float]
CoordDomain = tuple[CoordLimits, CoordLimits]
ExtentType = tuple[float, float, float, float]
SizeType = tuple[float, float]
StepType = tuple[float, float]

DEFAULT_STEP = (2.0, 2.0)


class RectangularRegion:

    domain: CoordDomain

    xticks: ArrayFloat32
    yticks: ArrayFloat32

    def __init__(self, domain: CoordDomain) -> None:
        # Validate domain to ensure increasing order and non-empty ranges
        self.domain = self._validate_domain(domain)

        self.xticks, self.yticks = self._create_grid_ticks(
            domain, DEFAULT_STEP
        )

    @classmethod
    def from_central_point(
        cls,
        size: SizeType,
        centre_deg: CenterType,
        units: Literal["arcsec", "arcmin", "deg", "rad"] = "deg",
    ) -> "RectangularRegion":
        size = cls._validate_size(size)
        size_deg = cls._convert_size_to_deg(size, units)
        centre_deg = cls._validate_centre(centre_deg)

        width_deg, height_deg = size_deg
        lon_cen, lat_cen = centre_deg

        x_half = 0.5 * width_deg
        y_half = 0.5 * height_deg

        lons_deg = (lon_cen - x_half, lon_cen + x_half)
        lats_deg = (lat_cen - y_half, lat_cen + y_half)

        domain = lons_deg, lats_deg

        return cls(domain)

    def set_ticks(
        self,
        domain: CoordDomain | None = None,
        step: StepType | float | int = DEFAULT_STEP,
    ) -> None:
        domain = (
            self.domain if domain is None else self._validate_domain(domain)
        )
        step = self._validate_step(step)
        self.xticks, self.yticks = self._create_grid_ticks(domain, step)

    @staticmethod
    def _convert_size_to_deg(size: SizeType, units: str) -> SizeType:
        # Validate units parameter
        allowed_units = {"arcsec", "arcmin", "deg", "rad"}
        if units not in allowed_units:
            raise ValueError(
                f"Invalid units: '{units}'; "
                f"allowed units are: {', '.join(allowed_units)}"
            )

        size_deg: tuple[float, ...]
        if units == "arcmin":
            size_deg = tuple(value / 60.0 for value in size)
        elif units == "arcsec":
            size_deg = tuple(value / 3600.0 for value in size)
        elif units == "rad":
            size_deg = tuple(degrees(value) for value in size)
        else:
            # units must be "deg" based on the validation
            size_deg = size

        return cast(SizeType, size_deg)

    @staticmethod
    def _create_grid_ticks(
        domain: CoordDomain, step: StepType
    ) -> tuple[ArrayFloat32, ArrayFloat32]:
        (lon_min, lon_max), (lat_min, lat_max) = domain
        lon_step, lat_step = step
        lon_min = lon_step * floor(lon_min / lon_step)
        lon_max = lon_step * floor(lon_max / lon_step) + lon_step
        xticks = arange(lon_min, lon_max, lon_step).astype(float32)

        lat_min = lat_step * floor(lat_min / lat_step)
        lat_max = lat_step * floor(lat_max / lat_step) + lat_step
        yticks = arange(lat_min, lat_max, lat_step).astype(float32)

        return cast(ArrayFloat32, xticks), cast(ArrayFloat32, yticks)

    @staticmethod
    def _validate_centre(centre_deg: CenterType) -> CenterType:
        # Validate 'centre_deg' structure, types, and range

        # Check if centre_deg is a tuple of two elements
        if not (isinstance(centre_deg, tuple) and len(centre_deg) == 2):
            raise TypeError(
                "'centre_deg' must be a tuple of two "
                "numbers (longitude, latitude)"
            )

        # Unpack the central point coordinates
        lon_deg, lat_deg = centre_deg

        # Check if the unpacked elements are numbers (int or float)
        if not (
            isinstance(lon_deg, (int, float))
            and isinstance(lat_deg, (int, float))
        ):
            raise TypeError(
                "Longitude and latitude values in 'centre_deg' must be numbers"
            )

        # Convert to float if they are integers
        try:
            lon_deg, lat_deg = float(lon_deg), float(lat_deg)
        except (ValueError, TypeError) as error:
            raise TypeError(
                "Longitude and latitude values in 'centre_deg' "
                "must be convertible to float"
            ) from error

        # Validate the geographical range for longitude and latitude in degrees
        # Assuming standard geographical ranges for degrees
        if not (-180.0 <= lon_deg <= 180.0):
            raise ValueError(
                f"Longitude ({lon_deg}) must be "
                "between -180.0 and 180.0 degrees"
            )

        if not (-90.0 <= lat_deg <= 90.0):
            raise ValueError(
                f"Latitude ({lat_deg}) must be "
                "between -90.0 and 90.0 degrees"
            )

        return lon_deg, lat_deg

    @staticmethod
    def _validate_size(size: SizeType) -> SizeType:
        # Validate 'size' structure, types, and range

        # Check if extent is a tuple of two elements
        if not (isinstance(size, tuple) and len(size) == 2):
            raise TypeError(
                "'size' must be a tuple of two numbers (width, height)"
            )

        width, height = size  # Unpack the extent values

        # Check if the unpacked elements are numbers (int or float)
        if not (
            isinstance(width, (int, float))
            and isinstance(height, (int, float))
        ):
            raise TypeError(
                "Width and height values in 'size' must be numbers"
            )

        # Convert to float if they are integers
        try:
            width, height = float(width), float(height)
        except (ValueError, TypeError) as error:
            raise TypeError(
                "Width and height values in 'size' "
                "must be convertible to float"
            ) from error

        # Validate that both values are positive
        if width <= 0 or height <= 0:
            raise ValueError(
                "'size' values must be positive, "
                f"received: ({width}, {height})"
            )

        return width, height

    @staticmethod
    def _validate_domain(domain: CoordDomain) -> CoordDomain:
        # Validate outer tuple structure
        if not (isinstance(domain, tuple) and len(domain) == 2):
            raise TypeError("'domain' must be a tuple of two 'CoordLimits'")

        # Unpack outer tuple (lon_limits, lat_limits)
        lon_limits, lat_limits = domain

        # Validate inner tuple structures and types of their elements
        if not (
            isinstance(lon_limits, tuple)
            and len(lon_limits) == 2
            and isinstance(lon_limits[0], (int, float))
            and isinstance(lon_limits[1], (int, float))
        ):
            raise TypeError(
                "The first element of 'domain' must be a tuple of two "
                "numbers (lon_min, lon_max)"
            )

        if not (
            isinstance(lat_limits, tuple)
            and len(lat_limits) == 2
            and isinstance(lat_limits[0], (int, float))
            and isinstance(lat_limits[1], (int, float))
        ):
            raise TypeError(
                "The second element of 'domain' must be a tuple of two "
                "numbers (lat_min, lat_max)"
            )

        # Unpack inner tuples (now that we know their structure is
        # likely correct)
        lon_min, lon_max = lon_limits
        lat_min, lat_max = lat_limits

        # Ensure values are floats after unpacking (optional, but good
        # practice if float is strictly required). This also handles
        # potential conversion if int was provided.
        try:
            lon_min, lon_max = float(lon_min), float(lon_max)
            lat_min, lat_max = float(lat_min), float(lat_max)
        except (ValueError, TypeError) as error:
            raise TypeError(
                "All 'domain' values must be numbers convertible to float"
            ) from error

        # Validate order
        if lon_min >= lon_max:
            raise ValueError(
                "Longitude values in 'domain' must be in increasing "
                "order (lon_min < lon_max)"
            )

        if lat_min >= lat_max:
            raise ValueError(
                "Latitude values in 'domain' must be in increasing "
                "order (lat_min < lat_max)"
            )

        return (lon_min, lon_max), (lat_min, lat_max)

    @staticmethod
    def _validate_step(step: StepType | float | int) -> StepType:
        # Ensure we have a tuple of values
        if not isinstance(step, tuple):
            step = (step, step)

        # Validate tuple structure and types of their elements
        if not (
            len(step) == 2
            and isinstance(step[0], (int, float))
            and isinstance(step[1], (int, float))
        ):
            raise TypeError(
                "'step' must be a number or a tuple of two "
                "numbers (lon_step, lat_step)"
            )

        # Unpack inner tuples (now that we know their structure is
        # likely correct)
        lon_step, lat_step = step

        # Ensure values are floats after unpacking (optional, but good
        # practice if float is strictly required). This also handles
        # potential conversion if int was provided.
        try:
            lon_step, lat_step = float(lon_step), float(lat_step)
        except (ValueError, TypeError) as error:
            raise TypeError(
                "All 'step' values must be numbers convertible to float"
            ) from error

        # Validate positiveness
        if lon_step > 0:
            raise ValueError(
                "Longitude increment in 'step' must be a positive number"
            )

        if lat_step > 0:
            raise ValueError(
                "Latitude increment in 'step' must be a positive number"
            )

        return lon_step, lat_step

    @property
    def extent(self) -> ExtentType:
        return self.lon_bounds + self.lat_bounds

    @property
    def lat_bounds(self) -> CoordLimits:
        return self.domain[1]

    @property
    def lon_bounds(self) -> CoordLimits:
        return self.domain[0]



import numpy as np
import numpy.fft as fft


def parabolic_interpolation(power_spectrum, peak_index):
    # Cannot interpolate across edges
    if peak_index <= 0 or peak_index >= len(power_spectrum) - 1:
        return 0, 0

    # Power values at three consecutive points
    y0 = power_spectrum[peak_index - 1]
    y1 = power_spectrum[peak_index]
    y2 = power_spectrum[peak_index + 1]

    # Avoid division by zero
    denominator = y0 - 2 * y1 + y2

    if np.abs(denominator) < 1e-10:
        return 0, 0

    numerator = y0 - y2

    # Calculating delta shift using parabolic interpolation formula
    delta = 0.5 * numerator / denominator

    true_index = peak_index + delta
    true_power = y1 - 0.25 * numerator * delta

    return delta, true_index, true_power


def find_frequencies2(
    signal,
    sampling_rate,
    num_freqs=1,
    apply_window=True,
    interp=True,
    min_amplitude=None,
):
    # Compute signal size
    signal_size = len(signal)

    # Apply window to reduce spectral leakage
    if apply_window:
        window = np.hanning(signal_size)
        signal = signal * window
        # Compensation factor
        scale = 1.0 / (window**2).mean()
    else:
        # Standard windowless normalization
        scale = 1.0 / signal_size

    # Calculate FFT and power spectrum
    fft_result = fft.rfft(signal)
    power_spectrum = np.abs(fft_result) ** 2 * scale

    # Peak detection with threshold (excluding DC component)
    all_indices = np.arange(1, len(power_spectrum))

    # Apply amplitude threshold if specified
    if min_amplitude is None:
        candidate_indices = all_indices
    else:
        valid_mask = power_spectrum[1:] >= min_amplitude
        candidate_indices = all_indices[valid_mask]

    # Find the indices of the strongest peaks (excluding DC component)
    ordered_indices = np.argsort(-power_spectrum[candidate_indices])
    sorted_indices = candidate_indices[ordered_indices]

    # Select dominant frequencies
    selected_indices = (
        sorted_indices[:num_freqs] if num_freqs else sorted_indices
    )

    # Get base frequencies
    frequencies = fft.rfftfreq(signal_size, d=1.0 / sampling_rate)

    # Case management without detected peaks
    if len(selected_indices) == 0:
        return np.array([]), frequencies, power_spectrum

    # Apply parabolic interpolation to each selected peak
    if interp:
        refined_freqs = []
        bin_resolution = sampling_rate / signal_size
        for idx in selected_indices:
            # Only interpolate valid peaks
            if 0 < idx < len(power_spectrum) - 1:
                delta = parabolic_interpolation(power_spectrum, idx)
                refined_freq = frequencies[idx] + delta * bin_resolution
                refined_freqs.append(refined_freq)
            else:
                refined_freqs.append(frequencies[idx])
        dominant_freqs = np.array(refined_freqs)
    else:
        dominant_freqs = frequencies[selected_indices]

    return dominant_freqs, frequencies, power_spectrum


def find_dominant_frequency(signal, sampling_rate, num_freqs):
    fft_result = fft.rfft(signal)
    frequencies = fft.rfftfreq(len(signal), d=1 / sampling_rate)
    idxs = np.argsort(np.abs(fft_result[1:]))[::-1][:num_freqs] + 1
    dominant_freqs = frequencies[idxs]
    dominant_freqs = dominant_freqs[dominant_freqs > 0]
    return dominant_freqs, frequencies, fft_result


def reduce_frequency_space(signal, n=3):
    fft_result = fft.fft(signal)
    fft_result[0] = 0
    fft_ordered = np.argsort(np.abs(fft_result))[::-1]
    fft_filtered = np.zeros_like(fft_result)
    fft_filtered[fft_ordered[:n]] = fft_result[fft_ordered[:n]]
    return fft.ifft(fft_filtered).real



"""
Provide methods for handling IR brightness temperature color enhancement.

It includes functions to load, parse, process, and save McIDAS and GMT
enhancement color tables, as well as reverse and manipulate color maps.
"""

ColorEntry = tuple[float, float, float, float]
ColorSegment = tuple[float, ...]
DomainData = tuple[float, float]

CMYKValue = ColorEntry
RGBValue = tuple[float, float, float]

ColorTable = list[RGBValue]
PaletteData = list[ColorEntry]
PaletteItem = tuple[PaletteData, ColorTable]
StretchingTable = list[tuple[float, float]]
ValueTables = tuple[list[float], ...]



"""
Extend the Datasource interface with cache and repository support.

Classes:
    DatasourceBase: Extend the Datasource interface.
"""

from .datasource import Datasource
from .datasource_cache import DatasourceCache


class DatasourceBase(Datasource):
    """
    Extend the Datasource interface with cache and repository support.

    Attributes
    ----------
    cache : DatasourceCache
        The cache for the datasource.
    """

    cache: DatasourceCache

    def __init__(
        self,
        base_url: str,
        cache: float | DatasourceCache | None,
    ) -> None:
        """
        Initialize the DatasourceBase.

        Parameters
        ----------
        base_url : str
            The base URL for the datasource.
        cache : float | DatasourceCache | None
            The cache for the datasource. If a float is provided, it
            will be used as the life time for each entry in the cache.
            If `None` is provided, the cache will be set to have a life
            time of 0.0 seconds, i.e. no caching.
        """
        super().__init__(base_url)

        if isinstance(cache, DatasourceCache):
            self.cache = cache
        else:
            self.cache = DatasourceCache(cache)



"""
Helper functions for the grid module.

These functions are used to calculate latitude and longitude grid data
from GOES ABI fixed grid projection data.

Functions
---------
make_consistent
    Make the latitude and longitude grids consistent.
make_common_mask
    Make the latitude and longitude grid masks consistent.
calculate_pixel_edges
    Calculate the coordinates of the edges of the pixels.
"""

from numpy import (
    concatenate,
    float64,
    isnan,
    nan,
    where,
)

from ..utils.array import ArrayFloat64


def make_consistent(
    abi_lon: ArrayFloat64, abi_lat: ArrayFloat64
) -> tuple[ArrayFloat64, ArrayFloat64]:
    """
    Make the latitude and longitude grids consistent.

    Helper function for functions calculate_latlon_grid_cartopy and
    calculate_latlon_grid_pyproj.

    Parameters
    ----------
    abi_lon : ArrayFloat64
        The longitude grid data.
    abi_lat : ArrayFloat64
        The latitude grid data.

    Returns
    -------
    tuple[ArrayFloat64, ArrayFloat64]
        A tuple containing the consistent latitude and longitude grid
        data.
    """
    valid_lon = (abi_lon >= -360.0) & (abi_lon <= 360.0)
    valid_lat = (abi_lat >= -90.0) & (abi_lat <= 90.0)
    is_valid = valid_lon & valid_lat

    abi_lon = where(is_valid, abi_lon, nan)
    abi_lat = where(is_valid, abi_lat, nan)

    abi_lon = where(abi_lon >= 180, abi_lon - 360, abi_lon)
    abi_lon = where(abi_lon < -180, abi_lon + 360, abi_lon)

    return abi_lon, abi_lat


def make_common_mask(
    abi_lat: ArrayFloat64, abi_lon: ArrayFloat64
) -> tuple[ArrayFloat64, ArrayFloat64]:
    """
    Make the latitude and longitude grid masks consistent.

    Helper function for function calculate_latlon_grid_noaa and
    compute_latlon_grid.

    Parameters
    ----------
    abi_lat : ArrayFloat64
        The latitude grid data.
    abi_lon : ArrayFloat64
        The longitude grid data.

    Returns
    -------
    tuple[ArrayFloat64, ArrayFloat64]
        A tuple containing consistently masked latitude and longitude
        grid data.
    """
    is_nan = isnan(abi_lat) | isnan(abi_lon)

    abi_lat = where(is_nan, nan, abi_lat)
    abi_lon = where(is_nan, nan, abi_lon)

    return abi_lat, abi_lon


def calculate_pixel_edges(centers: ArrayFloat64) -> ArrayFloat64:
    """
    Calculate the coordinates of the edges of the pixels.

    Given the coordinates of the center of pixels, calculate the
    coordinates of the edges of the pixels using linear interpolation,
    i.e. the midpoints between the centers. For the first and last
    elements, the edges are calculated using linear extrapolation.

    Parameters
    ----------
    centers : ArrayFloat64
        A n-elements 1-d array containing the coordinates of the center
        of pixels.

    Returns
    -------
    ArrayFloat64
        A (n+1)-elements 1-d array containing the coordinates of the
        edges of pixels.
    """
    left_offset: float64 = 2 * centers[0] - centers[1]
    right_offset: float64 = 2 * centers[-1] - centers[-2]

    left_centers = concatenate(([left_offset], centers))
    right_centers = concatenate((centers, [right_offset]))

    return 0.5 * (left_centers + right_centers)



from re import search

from ..netcdf import DatasetView, attribute
from .databook_gc import platform_origin_gridsat_gc


class PlatformMetadata(DatasetView):
    platform_domain: str = attribute("platform")

    @property
    def platform(self) -> str:
        return (
            match[0]
            if (match := search(r"GOES-\d{1,2}", self.platform_domain))
            else ""
        )

    @property
    def origin(self) -> str:
        return platform_origin_gridsat_gc[self.platform]
